#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_grant_access_control_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    access_controls = td.grant_access_control("foo", "bar", "baz", "hoge")
    td.post.assert_called_with("/v3/acl/grant", {"subject": "foo", "action": "bar", "scope": "baz", "grant_option": "hoge"})

def test_revoke_access_control_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    access_controls = td.revoke_access_control("foo", "bar", "baz")
    td.post.assert_called_with("/v3/acl/revoke", {"subject": "foo", "action": "bar", "scope": "baz"})

def test_test_access_control_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "permission": "",
            "access_controls": [
                {"subject":"foo","action":"bar","scope":"baz"},
                {"subject":"bar","action":"bar","scope":"baz"},
                {"subject":"baz","action":"bar","scope":"baz"}
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    access_controls = td.test_access_control("foo", "bar", "baz")
    td.get.assert_called_with("/v3/acl/test", {"user": "foo", "action": "bar", "scope": "baz"})

def test_list_access_controls_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "access_controls":[
              {"subject":"foo","action":"","scope":"","grant_option":""},
              {"subject":"bar","action":"","scope":"","grant_option":""},
              {"subject":"baz","action":"","scope":"","grant_option":""}
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    access_controls = td.list_access_controls()
    td.get.assert_called_with("/v3/acl/list")
    assert len(access_controls) == 3

def test_list_access_controls_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.list_access_controls()
    assert error.value.args == ("500: Listing access control failed: error",)



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class Database(Model):
    """Database on Treasure Data Service
    """

    PERMISSIONS = ["administrator", "full_access", "import_only", "query_only"]
    PERMISSION_LIST_TABLES = ["administrator", "full_access"]

    def __init__(self, client, db_name, tables=None, count=None, created_at=None, updated_at=None, org_name=None, permission=None):
        super(Database, self).__init__(client)
        self._db_name = db_name
        self._tables = tables
        self._count = count
        self._created_at = created_at
        self._updated_at = updated_at
        self._org_name = org_name
        self._permission = permission

    @property
    def org_name(self):
        """
        Returns: organization name
        """
        return self._org_name

    @property
    def permission(self):
        """
        Returns: a string represents permission for the database (e.g. "administrator", "full_access", etc.)
        """
        return self._permission

    @property
    def count(self):
        """
        TODO: add docstring
        """
        return self._count

    @property
    def name(self):
        """
        Returns: a name of the database in string
        """
        return self._db_name

    def tables(self):
        """
        Returns: a list of :class:`tdclient.model.Table`
        """
        if self._tables is None:
            self._update_tables()
        return self._tables

    def create_log_table(self, name):
        """
        Params:
            name (str): name of new log table

        Returns: :class:`tdclient.model.Table`
        """
        return self._client.create_log_table(self._db_name, name)

    def create_item_table(self, name):
        """
        Params:
            name (str): name of new item table

        Returns: :class:`tdclient.model.Table`
        """
        return self._client.create_item_table(self._db_name, name)

    def table(self, table_name):
        """
        Params:
            table_name (str): name of a table

        Returns: :class:`tdclient.model.Table`
        """
        return self._client.table(self._db_name, table_name)

    def delete(self):
        """Delete the database

        Returns: `True` if success
        """
        return self._client.delete_database(self._db_name)

    def query(self, q, **kwargs):
        """Run a query on the database

        Params:
            q (str): a query string

        Returns: :class:`tdclient.model.Job`
        """
        return self._client.query(self._db_name, q, **kwargs)

    @property
    def created_at(self):
        """
        Returns: :class:`datetime.datetime`
        """
        return self._created_at

    @property
    def updated_at(self):
        """
        Returns: :class:`datetime.datetime`
        """
        return self._updated_at

    def _update_tables(self):
        self._tables = self._client.tables(self._db_name)
        for table in self._tables:
            table.database = self



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_list_tables_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "tables":[
                {"id":210906,"name":"nasdaq","estimated_storage_size":168205061,"counter_updated_at":null,"last_log_timestamp":null,"type":"log","count":8807278,"expire_days":null,"created_at":"2014-10-08 02:57:38 UTC","updated_at":"2014-10-08 03:16:59 UTC","schema":"[[\\"symbol\\",\\"string\\"],[\\"open\\",\\"double\\"],[\\"volume\\",\\"long\\"],[\\"high\\",\\"double\\"],[\\"low\\",\\"double\\"],[\\"close\\",\\"double\\"]]"},
                {"id":208715,"name":"www_access","estimated_storage_size":0,"counter_updated_at":"2014-10-04T01:13:20Z","last_log_timestamp":"2014-10-04T01:13:15Z","type":"log","count":5000,"expire_days":null,"created_at":"2014-10-04 01:13:12 UTC","updated_at":"2014-10-22 07:02:19 UTC","schema":"[[\\"user\\",\\"int\\"],[\\"host\\",\\"string\\"],[\\"path\\",\\"string\\"],[\\"referer\\",\\"string\\"],[\\"code\\",\\"long\\"],[\\"agent\\",\\"string\\"],[\\"size\\",\\"long\\"],[\\"method\\",\\"string\\"]]"}
            ],
            "database":"sample_datasets"
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    tables = td.list_tables("sample_datasets")
    td.get.assert_called_with("/v3/table/list/sample_datasets")
    assert len(tables) == 2
    assert sorted(tables.keys()) == ["nasdaq", "www_access"]
    assert sorted([ v[0] for v in tables.values() ]) == ["log", "log"]

def test_list_tables_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.list_tables("sample_datasets")
    assert error.value.args == ("500: List tables failed: error",)

def test_create_log_table_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.create_log_table("sample_datasets", "nasdaq")
    td.post.assert_called_with("/v3/table/create/sample_datasets/nasdaq/log", {})

def test_create_item_table_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.create_item_table("sample_datasets", "nasdaq", "id", "INT")
    td.post.assert_called_with("/v3/table/create/sample_datasets/nasdaq/item", {"primary_key": "id", "primary_key_type": "INT"})

def test_swap_table_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.swap_table("sample_datasets", "foo", "bar")
    td.post.assert_called_with("/v3/table/swap/sample_datasets/foo/bar")

def test_update_schema_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.update_schema("sample_datasets", "foo", "{}")
    td.post.assert_called_with("/v3/table/update-schema/sample_datasets/foo", {"schema": "{}"})

def test_update_expire_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.update_expire("sample_datasets", "foo", 7)
    td.post.assert_called_with("/v3/table/update/sample_datasets/foo", {"expire_days": 7})

def test_delete_table_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "type": "item"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    td.delete_table("sample_datasets", "foo")
    td.post.assert_called_with("/v3/table/delete/sample_datasets/foo")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class Account(Model):
    """Account on Treasure Data Service
    """

    def __init__(self, client, account_id, plan, storage_size=None, guaranteed_cores=None, maximum_cores=None, created_at=None):
        super(Account, self).__init__(client)
        self._account_id = account_id
        self._plan = plan
        self._storage_size = storage_size
        self._guaranteed_cores = guaranteed_cores
        self._maximum_cores = maximum_cores
        self._created_at = created_at

    @property
    def account_id(self):
        """
        TODO: add docstring
        """
        return self._account_id

    @property
    def plan(self):
        """
        TODO: add docstring
        """
        return self._plan

    @property
    def storage_size(self):
        """
        TODO: add docstring
        """
        return self._storage_size

    @property
    def guaranteed_cores(self):
        """
        TODO: add docstring
        """
        return self._guaranteed_cores

    @property
    def maximum_cores(self):
        """
        TODO: add docstring
        """
        return self._maximum_cores

    @property
    def created_at(self):
        """
        TODO: add docstring
        """
        return self._created_at

    @property
    def storage_size_string(self):
        """
        TODO: add docstring
        """
        if self._storage_size <= 1024 * 1024:
            return "0.0 GB"
        elif self._storage_size <= 60 * 1024 * 1024:
            return "0.01 GB"
        elif self._storage_size <= 60 * 1024 * 1024 * 1024:
            return "%.1f GB" % (float(self._storage_size) / (1024 * 1024 * 1024))
        else:
            return "%d GB" % int(float(self._storage_size) / (1024 * 1024 * 1024))



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import io
import json
import msgpack
try:
    from unittest import mock
except ImportError:
    import mock
import pytest
import tempfile
import zlib

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_import_data_with_id_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "elapsed_time": "3.14"
        }
    """
    td.put = mock.MagicMock(return_value=make_response(200, body))
    elapsed_time = td.import_data("db", "table", "format", b"stream", 6, unique_id="unique_id")
    td.put.assert_called_with("/v3/table/import_with_id/db/table/unique_id/format", b"stream", 6)
    assert elapsed_time == 3.14

def test_import_data_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "elapsed_time": 2.71
        }
    """
    td.put = mock.MagicMock(return_value=make_response(200, body))
    elapsed_time = td.import_data("db", "table", "format", b"stream", 6)
    td.put.assert_called_with("/v3/table/import/db/table/format", b"stream", 6)
    assert elapsed_time == 2.71

def test_import_data_failure():
    td = api.API("APIKEY")
    td.put = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.import_data("db", "table", "format", b"stream", 6)
    assert error.value.args == ("500: Import failed: error",)

def msgpackb(list):
    """list -> bytes"""
    stream = io.BytesIO()
    packer = msgpack.Packer()
    for item in list:
        stream.write(packer.pack(item))
    return stream.getvalue()

def msgunpackb(bytes):
    """bytes -> list"""
    unpacker = msgpack.Unpacker(io.BytesIO(bytes), encoding=str("utf-8"))
    return list(unpacker)

def jsonb(list):
    """list -> bytes"""
    stream = io.BytesIO()
    for item in list:
        stream.write(json.dumps(item).encode("utf-8"))
        stream.write(b"\n")
    return stream.getvalue()

def unjsonb(bytes):
    """bytes -> list"""
    return [ json.loads(s.decode("utf-8")) for s in bytes.splitlines() ]
    return list(unpacker)

def gzipb(bytes):
    """bytes -> bytes"""
    compress = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
    return compress.compress(bytes) + compress.flush()

def gunzipb(bytes):
    """bytes -> bytes"""
    decompress = zlib.decompressobj(zlib.MAX_WBITS | 16)
    return decompress.decompress(bytes) + decompress.flush()

def test_import_file_msgpack_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    stream = io.BytesIO(msgpackb(data))
    td.import_file("db", "table", "msgpack", stream)

def test_import_file_msgpack_file_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    # should not use `tempfile.NamedTemporaryFile` to fix tests working on Windows
    # http://bugs.python.org/issue14243
    name = None
    try:
        fd, name = tempfile.mkstemp("wb")
        fp = os.fdopen(fd, "wb")
        fp.write(msgpackb(data))
        fp.close()
        td.import_file("db", "table", "msgpack", name)
    finally:
        if name is not None:
            os.unlink(name)

def test_import_file_msgpack_failure():
    td = api.API("APIKEY")
    td.import_data = mock.MagicMock()
    stream = io.BytesIO(b"\xc1 malformed msgpack")
    with pytest.raises(ValueError) as error:
        td.import_file("db", "table", "msgpack", stream)

def test_import_file_msgpack_gz_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    stream = io.BytesIO(gzipb(msgpackb(data)))
    td.import_file("db", "table", "msgpack.gz", stream)

def test_import_file_msgpack_gz_file_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    # should not use `tempfile.NamedTemporaryFile` to fix tests working on Windows
    # http://bugs.python.org/issue14243
    name = None
    try:
        fd, name = tempfile.mkstemp("wb")
        fp = os.fdopen(fd, "wb")
        fp.write(gzipb(msgpackb(data)))
        fp.close()
        td.import_file("db", "table", "msgpack.gz", name)
    finally:
        if name is not None:
            os.unlink(name)

def test_import_file_msgpack_gz_failure():
    td = api.API("APIKEY")
    td.import_data = mock.MagicMock()
    stream = io.BytesIO(b"\xc1 malformed msgpack.gz")
    with pytest.raises(IOError) as error:
        td.import_file("db", "table", "msgpack.gz", stream)

def test_import_file_json_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    stream = io.BytesIO(jsonb(data))
    td.import_file("db", "table", "json", stream)

def test_import_file_json_failure():
    td = api.API("APIKEY")
    td.import_data = mock.MagicMock()
    stream = io.BytesIO(b"malformed json")
    with pytest.raises(ValueError) as error:
        td.import_file("db", "table", "json", stream)

def test_import_file_json_gz_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    def import_data(db, table, format, stream, size, unique_id=None):
        assert db == "db"
        assert table == "table"
        assert format == "msgpack.gz"
        assert msgunpackb(gunzipb(stream.read(size))) == data
        assert unique_id is None
    td.import_data = import_data
    stream = io.BytesIO(gzipb(jsonb(data)))
    td.import_file("db", "table", "json.gz", stream)

def test_import_file_json_gz_failure():
    td = api.API("APIKEY")
    td.import_data = mock.MagicMock()
    stream = io.BytesIO(b"malformed json.gz")
    with pytest.raises(IOError) as error:
        td.import_file("db", "table", "json.gz", stream)

def test_import_file_unknown_format():
    td = api.API("APIKEY")
    td.import_data = mock.MagicMock()
    stream = io.BytesIO(b"malformed json.gz")
    with pytest.raises(TypeError) as error:
        td.import_file("db", "table", "UNKNOWN", stream)



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class User(Model):
    """User on Treasure Data Service
    """

    def __init__(self, client, name, org_name, role_names, email):
        super(User, self).__init__(client)
        self._name = name
        self._org_name = org_name
        self._role_names = role_names
        self._email = email

    @property
    def name(self):
        """
        Returns: name of the user
        """
        return self._name

    @property
    def org_name(self):
        """
        Returns: organization name
        """
        return self._org_name

    @property
    def role_names(self):
        """
        TODO: add docstring
        """
        return self._role_names

    @property
    def email(self):
        """
        Returns: e-mail address
        """
        return self._email



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_user():
    client = mock.MagicMock()
    user = model.User(client, "name", "org_name", ["role1", "role2"], "email")
    assert user.name == "name"
    assert user.org_name == "org_name"
    assert user.role_names == ["role1", "role2"]
    assert user.email == "email"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from urllib.parse import quote as urlquote # >=3.0
except ImportError:
    from urllib import quote as urlquote

class UserAPI(object):
    ####
    ## User API
    ##

    def authenticate(self, user, password):
        """
        TODO: add docstring
        apikey:str
        """
        with self.post("/v3/user/authenticate", {"user": user, "password": password}) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Authentication failed", res, body)
            js = self.checked_json(body, ["apikey"])
            apikey = js["apikey"]
            return apikey

    def list_users(self):
        """
        TODO: add docstring
        => [[name:str,organization:str,[user:str]]
        """
        with self.get("/v3/user/list") as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List users failed", res, body)
            js = self.checked_json(body, ["users"])
            def user(roleinfo):
                name = roleinfo["name"]
                email = roleinfo["email"]
                return (name, None, None, email) # set None to org and role for API compatibility
            return [ user(roleinfo) for roleinfo in js["users"] ]

    def add_user(self, name, org, email, password):
        """
        TODO: add docstring
        => True
        """
        params = {"organization": org, "email": email, "password": password}
        with self.post("/v3/user/add/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Adding user failed", res, body)
            return True

    def remove_user(self, name):
        """
        TODO: add docstring
        => True
        """
        with self.post("/v3/user/remove/%s" % (urlquote(str(name)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Removing user failed", res, body)
            return True

    def change_email(self, name, email):
        """
        TODO: add docstring
        => True
        """
        params = {"email": email}
        with self.post("/v3/user/email/change/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Changing email failed", res, body)
            return True

    def list_apikeys(self, name):
        """
        TODO: add docstring
        => [apikey:str]
        """
        with self.get("/v3/user/apikey/list/%s" % (urlquote(str(name)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List API keys failed", res, body)
            js = self.checked_json(body, ["apikeys"])
            return js["apikeys"]

    def add_apikey(self, name):
        """
        TODO: add docstring
        => True
        """
        with self.post("/v3/user/apikey/add/%s" % (urlquote(str(name)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Adding API key failed", res, body)
            return True

    def remove_apikey(self, name, apikey):
        """
        TODO: add docstring
        => True
        """
        params = {"apikey": apikey}
        with self.post("/v3/user/apikey/remove/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Removing API key failed", res, body)
            return True

    def change_password(self, name, password):
        """
        TODO: add docstring
        => True
        """
        params = {"password": password}
        with self.post("/v3/user/password/change/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Changing password failed", res, body)
            return True

    def change_my_password(self, old_password, password):
        """
        TODO: add docstring
        => True
        """
        params = {"old_password": old_password, "password": password}
        with self.post("/v3/user/password/change", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Changing password failed", res, body)
            return True



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import time

class AccountAPI(object):
    ####
    ## Account API
    ##

    def show_account(self):
        """
        TODO: add docstring
        """
        with self.get("/v3/account/show") as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Show account failed", res, body)
            js = self.checked_json(body, ["account"])
            a = js["account"]
            account_id = int(a["id"])
            plan = int(a["plan"])
            storage_size = int(a["storage_size"])
            guaranteed_cores = int(a["guaranteed_cores"])
            maximum_cores = int(a["maximum_cores"])
            created_at = self.parsedate(a["created_at"])
            return [account_id, plan, storage_size, guaranteed_cores, maximum_cores, created_at]

    def account_core_utilization(self, _from, to):
        """
        TODO: add docstring
        """
        params = {}
        if _from is not None:
            params["from"] = str(_from)
        if to is not None:
            params["to"] = str(to)
        with self.get("/v3/account/core_utilization", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Show account failed", res, body)
            js = self.checked_json(body, ["from", "to", "interval", "history"])
            _from = self.parsedate(js["from"])
            to = self.parsedate(js["to"])
            interval = int(js["interval"])
            history = js["history"]
            return [_from, to, interval, history]



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from urllib.parse import quote as urlquote # >=3.0
except ImportError:
    from urllib import quote as urlquote

class ScheduleAPI(object):
    ####
    ## Schedule API
    ##

    def create_schedule(self, name, params={}):
        """
        TODO: add docstring
        => start:str
        """
        params.update({"type": params.get("type", "hive")})
        with self.post("/v3/schedule/create/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Create schedule failed", res, body)
            js = self.checked_json(body, ["start"])
            return self.parsedate(js["start"])

    def delete_schedule(self, name):
        """
        TODO: add docstring
        => cron:str, query:str
        """
        with self.post("/v3/schedule/delete/%s" % (urlquote(str(name)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Delete schedule failed", res, body)
            js = self.checked_json(body, ["cron", "query"])
            return (js["cron"], js["query"])

    def list_schedules(self):
        """
        TODO: add docstring
        => [(name:str, cron:str, query:str, database:str, result_url:str)]
        """
        with self.get("/v3/schedule/list") as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List schedules failed", res, body)
            js = self.checked_json(body, ["schedules"])
            def schedule(m):
                name = m.get("name")
                cron = m.get("cron")
                query = m.get("query")
                database = m.get("database")
                result_url = m.get("result")
                timezone = m.get("timezone", "UTC")
                delay = m.get("delay")
                next_time = self.parsedate(self.get_or_else(m, "next_time", "1970-01-01T00:00:00Z"))
                priority = m.get("priority")
                retry_limit = m.get("retry_limit")
                return (name, cron, query, database, result_url, timezone, delay, next_time, priority, retry_limit, None) # same as database
            return [ schedule(m) for m in js["schedules"] ]

    def update_schedule(self, name, params={}):
        """
        TODO: add docstring
        """
        with self.post("/v3/schedule/update/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Update schedule failed", res, body)

    def history(self, name, _from=0, to=None):
        """
        TODO: add docstring
        """
        params = {}
        if _from is not None:
            params["from"] = str(_from)
        if to is not None:
            params["to"] = str(to)
        with self.get("/v3/schedule/history/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List history failed", res, body)
            js = self.checked_json(body, ["history"])
            def history(m):
                job_id = m.get("job_id")
                type = m.get("type", "?")
                database = m.get("database")
                status = m.get("status")
                query = m.get("query")
                start_at = self.parsedate(self.get_or_else(m, "start_at", "1970-01-01T00:00:00Z"))
                end_at = self.parsedate(self.get_or_else(m, "end_at", "1970-01-01T00:00:00Z"))
                scheduled_at = self.parsedate(self.get_or_else(m, "scheduled_at", "1970-01-01T00:00:00Z"))
                result_url = m.get("result")
                priority = m.get("priority")
                return (scheduled_at, job_id, type, status, query, start_at, end_at, result_url, priority, database)
            return [ history(m) for m in js["history"] ]

    def run_schedule(self, name, time, num=None):
        """
        TODO: add docstring
        """
        params = {}
        if num is not None:
            params = {"num": num}
        with self.post("/v3/schedule/run/%s/%s" % (urlquote(str(name)), urlquote(str(time))), params) as res:
            code, body = res.status, res.read()
        if code != 200:
            self.raise_error("Run schedule failed", res, body)
        js = self.checked_json(body, ["jobs"])
        def job(m):
            job_id = m.get("job_id")
            scheduled_at = self.parsedate(self.get_or_else(m, "scheduled_at", "1970-01-01T00:00:00Z"))
            type = m.get("type", "?")
            return (job_id, type, scheduled_at)
        return [ job(m) for m in js["jobs"] ]



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import json

from tdclient import api
from tdclient import model

class Client(object):
    """API Client for Treasure Data Service
    """

    def __init__(self, *args, **kwargs):
        self._api = api.API(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @property
    def api(self):
        """
        Returns: an instance of :class:`tdclient.api.API`
        """
        return self._api

    @property
    def apikey(self):
        """
        Returns: API key string.
        """
        return self._api.apikey

    def server_status(self):
        """
        Returns: a string represents current server status.
        """
        return self.api.server_status()

    def create_database(self, db_name, **kwargs):
        """
        Params:
            db_name (str): name of a database to create

        Returns: `True` if success
        """
        return self.api.create_database(db_name, **kwargs)

    def delete_database(self, db_name):
        """
        Params:
            db_name (str): name of database to delete

        Returns: `True` if success
        """
        return self.api.delete_database(db_name)

    def account(self):
        """
        Returns: :class:`tdclient.model.Acount`
        """
        account_id, plan, storage, guaranteed_cores, maximum_cores, created_at = self.api.show_account()
        return model.Account(self, account_id, plan, storage, guaranteed_cores, maximum_cores, created_at)

    def core_utilization(self, _from, to):
        """
        TODO: add docstring
        """
        _from, to, interval, history = self.api.account_core_utilization(_from, to)
        return (_from, to, interval, history)

    def databases(self):
        """
        Returns: a list of :class:`tdclient.model.Database`
        """
        databases = self.api.list_databases()
        return [ model.Database(self, db_name, None, *args) for (db_name, args) in databases.items() ]

    def database(self, db_name):
        """
        Params:
            db_name (str): name of a database

        Returns: :class:`tdclient.model.Database`
        """
        databases = self.api.list_databases()
        for (name, args) in databases.items():
            if name == db_name:
                return model.Database(self, name, None, *args)
        raise api.NotFoundError("Database '%s' does not exist" % (db_name))

    def create_log_table(self, db_name, table_name):
        """
        Params:
            db_name (str): name of a database
            table_name (str): name of a table to create

        Returns: `True` if success
        """
        return self.api.create_log_table(db_name, table_name)

    def create_item_table(self, db_name, table_name, primary_key, primary_key_type):
        """
        Params:
            db_name (str): name of a database
            table_name (str): name of a table to create
            primary_key (str): name of primary key column
            primary_key_type (str): type of primary key column

        Returns: `True` if success
        """
        return self.api.create_item_table(db_name, table_name, primary_key, primary_key_type)

    def swap_table(self, db_name, table_name1, table_name2):
        """
        Params:
            db_name (str): name of a database
            table_name1 (str): original table name
            table_name2 (str): table name you want to rename to

        Returns: `True` if success
        """
        return self.api.swap_table(db_name, table_name1, table_name2)

    def update_schema(self, db_name, table_name, schema):
        """Updates the schema of a table

        Params:
            db_name (str): name of a database
            table_name (str): name of a table
            schema (dict): a dictionary object represents the schema definition (will convert to JSON)

        Returns: `True` if success
        """
        return self.api.update_schema(db_name, table_name, json.dumps(schema))

    def update_expire(self, db_name, table_name, expire_days):
        """Set expiration date to a table

        Params:
            db_name (str): name of a database
            table_name (str): name of a table
            epire_days (int): expiration date in days from today

        Returns: `True` if success
        """
        return self.api.update_expire(db_name, table_name, expire_days)

    def delete_table(self, db_name, table_name):
        """Delete a table

        Params:
            db_name (str): name of a database
            table_name (str): name of a table

        Returns: a string represents the type of deleted table
        """
        return self.api.delete_table(db_name, table_name)

    def tables(self, db_name):
        """List existing tables

        Params:
            db_name (str): name of a database

        Returns: a list of :class:`tdclient.model.Table`
        """
        m = self.api.list_tables(db_name)
        return [ model.Table(self, db_name, table_name, *args) for (table_name, args) in m.items() ]

    def table(self, db_name, table_name):
        """
        Params:
            db_name (str): name of a database
            table_name (str): name of a table

        Returns: :class:`tdclient.model.Table`

        Raises:
            tdclient.api.NotFoundError: if the table doesn't exist
        """
        tables = self.tables(db_name)
        for table in tables:
            if table.table_name == table_name:
                return table
        raise api.NotFoundError("Table '%s.%s' does not exist" % (db_name, table_name))

    def tail(self, db_name, table_name, count, to=None, _from=None, block=None):
        """
        TODO: add docstring
        """
        return self.api.tail(db_name, table_name, count, to, _from, block)

    def query(self, db_name, q, result_url=None, priority=None, retry_limit=None, type="hive", **kwargs):
        """Run a query on specified database table.

        Params:
            db_name (str): name of a database
            q (str): a query string
            result_url (str): result output URL
            priority (int or str): priority (e.g. "NORMAL", "HIGH", etc.)
            retry_limit (int): retry limit
            type (str): name of a query engine

        Returns: :class:`tdclient.model.Job`

        Raises:
            ValueError: if unknown query type has been specified
        """
        # for compatibility, assume type is hive unless specifically specified
        if type not in ["hive", "pig", "impala", "presto"]:
            raise ValueError("The specified query type is not supported: %s" % (type))
        job_id = self.api.query(q, type=type, db=db_name, result_url=result_url, priority=priority, retry_limit=retry_limit, **kwargs)
        return model.Job(self, job_id, type, q)

    def jobs(self, _from=None, to=None, status=None, conditions=None):
        """List jobs

        Params:
            _from (int):
            to (int):
            status (str):
            conditions (str)

        Returns: a list of :class:`tdclient.model.Job`
        """
        results = self.api.list_jobs(_from, to, status, conditions)
        def job(d):
            return model.Job(
                self,
                d["job_id"],
                d["type"],
                d["query"],
                status=d.get("status"),
                url=d.get("url"),
                debug=d.get("debug"),
                start_at=d.get("start_at"),
                end_at=d.get("end_at"),
                cpu_time=d.get("cpu_time"),
                result_size=d.get("result_size"),
                result=d.get("result"),
                result_url=d.get("result_url"),
                hive_result_schema=d.get("hive_result_schema"),
                priority=d.get("priority"),
                retry_limit=d.get("retry_limit"),
                org_name=d.get("org_name"),
                db_name=d.get("db_name"),
            )
        return [ job(d) for d in results ]

    def job(self, job_id):
        """Get a job from `job_id`

        Params:
            job_id (str): job id

        Returns: :class:`tdclient.model.Job`
        """
        d = self.api.show_job(str(job_id))
        return model.Job(
            self,
            job_id,
            d["type"],
            d["query"],
            status=d.get("status"),
            url=d.get("url"),
            debug=d.get("debug"),
            start_at=d.get("start_at"),
            end_at=d.get("end_at"),
            cpu_time=d.get("cpu_time"),
            result_size=d.get("result_size"),
            result=d.get("result"),
            result_url=d.get("result_url"),
            hive_result_schema=d.get("hive_result_schema"),
            priority=d.get("priority"),
            retry_limit=d.get("retry_limit"),
            org_name=d.get("org_name"),
            db_name=d.get("db_name"),
        )

    def job_status(self, job_id):
        """
        Params:
            job_id (str): job id

        Returns: a string represents the status of the job ("success", "error", "killed", "queued", "running")
        """
        return self.api.job_status(job_id)

    def job_result(self, job_id):
        """
        Params:
            job_id (str): job id

        Returns: a list of each rows in result set
        """
        return self.api.job_result(job_id)

    def job_result_each(self, job_id):
        """
        Params:
            job_id (str): job id

        Returns: an iterator of result set
        """
        for row in self.api.job_result_each(job_id):
            yield row

    def job_result_format(self, job_id, format):
        """
        Params:
            job_id (str): job id
            format (str): output format of result set

        Returns: a list of each rows in result set
        """
        return self.api.job_result_format(job_id, format)

    def job_result_format_each(self, job_id, format):
        """
        Params:
            job_id (str): job id
            format (str): output format of result set

        Returns: an iterator of rows in result set
        """
        for row in self.api.job_result_format_each(job_id, format):
            yield row

    def kill(self, job_id):
        """
        Params:
            job_id (str): job id

        Returns: a string represents the status of killed job ("queued", "running")
        """
        return self.api.kill(job_id)

    def export_data(self, db_name, table_name, storage_type, params={}):
        """Export data from Treasure Data Service

        Params:
            db_name (str): name of a database
            table_name (str): name of a table
            storage_type (str): type of the storage
            params (dict): optional parameters

        Returns: :class:`tdclient.model.Job`
        """
        job_id = self.api.export_data(db_name, table_name, storage_type, params)
        return model.Job(self, job_id, "export", None)

    def partial_delete(self, db_name, table_name, to, _from, params={}):
        """
        TODO: add docstring
        => :class:`tdclient.model.Job`
        """
        job_id = self.api.partial_delete(db_name, table_name, to, _from, params)
        return model.Job(self, job_id, "partialdelete", None)

    def create_bulk_import(self, name, database, table, params={}):
        """Create new bulk import session

        Params:
            name (str): name of new bulk import session
            database (str): name of a database
            table (str): name of a table
        """
        return self.api.create_bulk_import(name, database, table, params)

    def delete_bulk_import(self, name):
        """Delete a bulk import session

        Params:
            name (str): name of a bulk import session
        """
        return self.api.delete_bulk_import(name)

    def freeze_bulk_import(self, name):
        """Freeze a bulk import session

        Params:
            name (str): name of a bulk import session
        """
        return self.api.freeze_bulk_import(name)

    def unfreeze_bulk_import(self, name):
        """Unfreeze a bulk import session

        Params:
            name (str): name of a bulk import session
        """
        return self.api.unfreeze_bulk_import(name)

    def perform_bulk_import(self, name):
        """Perform a bulk import session

        Params:
            name (str): name of a bulk import session

        Returns: :class:`tdclient.model.Job`
        """
        job_id = self.api.perform_bulk_import(name)
        return model.Job(self, job_id, "bulk_import", None)

    def commit_bulk_import(self, name):
        """Commit a bulk import session

        Params:
            name (str): name of a bulk import session
        """
        return self.api.commit_bulk_import(name)

    def bulk_import_error_records(self, name):
        """
        Params:
            name (str): name of a bulk import session

        Returns: an iterator of error records
        """
        for record in self.api.bulk_import_error_records(name):
            yield record

    def bulk_import(self, name):
        """Get a bulk import session

        Params:
            name (str): name of a bulk import session

        Returns: :class:`tdclient.model.BulkImport`
        """
        data = self.api.show_bulk_import(name)
        return model.BulkImport(self, **data)

    def bulk_imports(self):
        """List bulk import sessions

        Returns: a list of :class:`tdclient.model.BulkImport`
        """
        return [ model.BulkImport(self, **data) for data in self.api.list_bulk_imports() ]

    def bulk_import_upload_part(self, name, part_name, stream, size):
        """Upload a part to a bulk import session

        Params:
            name (str): name of a bulk import session
            part_name (str): name of a part of the bulk import session
            stream (file-like): a file-like object contains the part
            size (int): the size of the part
        """
        return self.api.bulk_import_upload_part(name, part_name, stream, size)

    def bulk_import_delete_part(self, name, part_name):
        """Delete a part from a bulk import session

        Params:
            name (str): name of a bulk import session
            part_name (str): name of a part of the bulk import session
        """
        return self.api.bulk_import_delete_part(name, part_name)

    def list_bulk_import_parts(self, name):
        """List parts of a bulk import session

        Params:
            name (str): name of a bulk import session

        Returns: a list of string represents the name of parts
        """
        return self.api.list_bulk_import_parts(name)

    def create_schedule(self, name, params={}):
        """
        TODO: add docstring
        => first_time:datetime.datetime
        """
        if "cron" not in params:
            raise ValueError("'cron' option is required")
        if "query" not in params:
            raise ValueError("'query' option is required")
        return self.api.create_schedule(name, params)

    def delete_schedule(self, name):
        """
        TODO: add docstring
        => True
        """
        return self.api.delete_schedule(name)

    def schedules(self):
        """
        TODO: add docstring
        [:class:`tdclient.model.Schedule`]
        """
        result = self.api.list_schedules()
        def schedule(m):
            name,cron,query,database,result_url,timezone,delay,next_time,priority,retry_limit,org_name = m
            return model.Schedule(self, name, cron, query, database, result_url, timezone, delay, next_time, priority, retry_limit, org_name)
        return [ schedule(m) for m in result ]

    def update_schedule(self, name, params={}):
        """
        TODO: add docstring
        [:class:`tdclient.model.ScheduledJob`]
        """
        self.api.update_schedule(name, params)

    def history(self, name, _from=None, to=None):
        """
        TODO: add docstring
        [:class:`tdclient.model.ScheduledJob`]
        """
        result = self.api.history(name, _from, to)
        def scheduled_job(m):
            scheduled_at,job_id,type,status,query,start_at,end_at,result_url,priority,database = m
            job_param = [job_id, type, query, status,
                None, None, # url, debug
                start_at, end_at,
                None, # cpu_time
                None, None, # result_size, result
                result_url,
                None, # hive_result_schema
                priority,
                None, # retry_limit
                None, # TODO org_name
                database]
            return model.ScheduledJob(self, scheduled_at, *job_param)
        return [ scheduled_job(m) for m in result ]

    def run_schedule(self, name, time, num):
        """
        TODO: add docstring
        [:class:`tdclient.model.ScheduledJob`]
        """
        results = self.api.run_schedule(name, time, num)
        def scheduled_job(m):
            job_id,type,scheduled_at = m
            return model.ScheduledJob(self, scheduled_at, job_id, type, None)
        return [ scheduled_job(m) for m in results ]

    def import_data(self, db_name, table_name, format, bytes_or_stream, size, unique_id=None):
        """Import data into Treasure Data Service

        Params:
            db_name (str): name of a database
            table_name (str): name of a table
            format (str): format of data type (e.g. "msgpack.gz")
            bytes_or_stream (str or file-like): a byte string or a file-like object contains the data
            size (int): the length of the data
            unique_id (str): a unique identifier of the data

        Returns: second in float represents elapsed time to import data
        """
        return self.api.import_data(db_name, table_name, format, bytes_or_stream, size, unique_id=unique_id)

    def import_file(self, db_name, table_name, format, file, unique_id=None):
        """Import data into Treasure Data Service, from an existing file on filesystem.

        This method will decompress/deserialize records from given file, and then
        convert it into format acceptable from Treasure Data Service ("msgpack.gz").

        Params:
            db (str): name of a database
            table (str): name of a table
            format (str): format of data type (e.g. "msgpack", "json")
            file (str or file-like): a name of a file, or a file-like object contains the data
            unique_id (str): a unique identifier of the data

        Returns: float represents the elapsed time to import data
        """
        return self.api.import_file(db_name, table_name, format, file, unique_id=unique_id)

    def results(self):
        """
        Returns: a list of :class:`tdclient.model.Result`
        """
        results = self.api.list_result()
        def result(m):
            name,url,organizations = m
            return model.Result(self, name, url, organizations)
        return [ result(m) for m in results ]

    def create_result(self, name, url, params={}):
        """
        TODO: add docstring
        => True
        """
        return self.api.create_result(name, url, params)

    def delete_result(self, name):
        """
        TODO: add docstring
        => True
        """
        return self.api.delete_result(name)

    def users(self):
        """List users

        Returns: a liast of :class:`tdclient.model.User`
        """
        results = self.api.list_users()
        def user(m):
            name,org,roles,email = m
            return model.User(self, name, org, roles, email)
        return [ user(m) for m in results ]

    def add_user(self, name, org, email, password):
        """Add a new user

        Params:
            name (str): name of the user
            org (str): organization
            email: (str): e-mail address
            password (str): password

        Returns: `True` if success
        """
        return self.api.add_user(name, org, email, password)

    def remove_user(self, name):
        """Remove a user

        Params:
            name (str): name of the user

        Returns: `True` if success
        """
        return self.api.remove_user(name)

    def change_email(self, name, email):
        """
        Params:
            name (str): name of the user
            email (str) new e-mail address

        Returns: `True` if success
        """
        return self.api.change_email(name, email)

    def list_apikeys(self, name):
        """
        Params:
            name (str): name of the user

        Returns: a list of string of API key
        """
        return self.api.list_apikeys(name)

    def add_apikey(self, name):
        """
        Params:
            name (str): name of the user

        Returns: `True` if success
        """
        return self.api.add_apikey(name)

    def remove_apikey(self, name, apikey):
        """
        Params:
            name (str): name of the user
            apikey (str): an API key to remove

        Returns: `True` if success
        """
        return self.api.remove_apikey(name, apikey)

    def change_password(self, name, password):
        """
        Params:
            name (str): name of the user
            password (str): new password

        Returns: `True` if success
        """
        return self.api.change_password(name, password)

    def change_my_password(self, old_password, password):
        """
        Params:
            old_password (str): old password
            password (str): new password

        Returns: `True` if success
        """
        return self.api.change_my_password(old_password, password)

    def access_controls(self):
        """
        Returns: a list of :class:`tdclient.model.AccessControl`
        """
        results = self.api.list_access_controls()
        def access_control(m):
            subject,action,scope,grant_option = m
            return model.AccessControl(self, subject, action, scope, grant_option)
        return [ access_control(m) for m in results ]

    def grant_access_control(self, subject, action, scope, grant_option):
        """
        TODO: add docstring
        => True
        """
        return self.api.grant_access_control(subject, action, scope, grant_option)

    def revoke_access_control(self, subject, action, scope):
        """
        TODO: add docstring
        => True
        """
        return self.api.revoke_access_control(subject, action, scope)

    def test_access_control(self, user, action, scope):
        """
        TODO: add docstring
        => True
        """
        return self.api.test_access_control(user, action, scope)

    def close(self):
        """Close opened API connections.
        """
        return self._api.close()



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import datetime
import dateutil.tz
try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_schema():
    client = mock.MagicMock()
    job = model.Job(client, "job_id", "type", "query", status="status", url="url", debug="debug", start_at="start_at", end_at="end_at", cpu_time="cpu_time", result_size="result_size", result="result", result_url="result_url", hive_result_schema="hive_result_schema", priority="UNKNOWN", retry_limit="retry_limit", org_name="org_name", db_name="db_name")
    assert job.id == "job_id"
    assert job.job_id == "job_id"
    assert job.type == "type"
    assert job.result_url == "result_url"
    assert job.priority == "UNKNOWN"
    assert job.retry_limit == "retry_limit"
    assert job.org_name == "org_name"
    assert job.db_name == "db_name"

def test_job_priority():
    client = mock.MagicMock()
    assert model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", priority=-2).priority == "VERY LOW"
    assert model.Job(client, "2", "hive", "SELECT COUNT(1) FROM nasdaq", priority=-1).priority == "LOW"
    assert model.Job(client, "3", "hive", "SELECT COUNT(1) FROM nasdaq", priority=0).priority == "NORMAL"
    assert model.Job(client, "4", "hive", "SELECT COUNT(1) FROM nasdaq", priority=1).priority == "HIGH"
    assert model.Job(client, "5", "hive", "SELECT COUNT(1) FROM nasdaq", priority=2).priority == "VERY HIGH"
    assert model.Job(client, "42", "hive", "SELECT COUNT(1) FROM nasdaq", priority=42).priority == "42"

def test_job_wait_success():
    client = mock.MagicMock()
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(side_effect=[
        False,
        True,
    ])
    with mock.patch("time.time") as t_time:
        t_time.side_effect = [
            1423570800.0,
            1423570860.0,
            1423570920.0,
            1423570980.0,
        ]
        with mock.patch("time.sleep") as t_sleep:
            job.wait(timeout=120)
            assert t_sleep.called
        assert t_time.called
    assert job.finished.called

def test_job_wait_failure():
    client = mock.MagicMock()
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=False)
    with mock.patch("time.time") as t_time:
        t_time.side_effect = [
            1423570800.0,
            1423570860.0,
            1423570920.0,
            1423570980.0,
        ]
        with mock.patch("time.sleep") as t_sleep:
            with pytest.raises(RuntimeError) as error:
                job.wait(timeout=120)
                assert t_sleep.called
        assert t_time.called
    assert job.finished.called

def test_job_kill():
    client = mock.MagicMock()
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.kill()
    client.kill.assert_called_with("12345")

def test_job_result_generator():
    client = mock.MagicMock()
    def job_result_each(job_id):
        assert job_id == "12345"
        yield ["foo", 123]
        yield ["bar", 456]
        yield ["baz", 789]
    client.job_result_each = job_result_each
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=True)
    rows = []
    for row in job.result():
        rows.append(row)
    assert rows == [["foo", 123], ["bar", 456], ["baz", 789]]

def test_job_result_list():
    client = mock.MagicMock()
    result = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq", result=result)
    job.finished = mock.MagicMock(return_value=True)
    rows = []
    for row in job.result():
        rows.append(row)
    assert rows == [["foo", 123], ["bar", 456], ["baz", 789]]

def test_job_result_failure():
    client = mock.MagicMock()
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=False)
    with pytest.raises(ValueError) as error:
        for row in job.result():
            pass

def test_job_result_format_generator():
    client = mock.MagicMock()
    def job_result_format_each(job_id, format):
        assert job_id == "12345"
        assert format == "msgpack.gz"
        yield ["foo", 123]
        yield ["bar", 456]
        yield ["baz", 789]
    client.job_result_format_each = job_result_format_each
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=True)
    rows = []
    for row in job.result_format("msgpack.gz"):
        rows.append(row)
    assert rows == [["foo", 123], ["bar", 456], ["baz", 789]]

def test_job_result_format_list():
    client = mock.MagicMock()
    result = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq", result=result)
    job.finished = mock.MagicMock(return_value=True)
    rows = []
    for row in job.result_format("msgpack.gz"):
        rows.append(row)
    assert rows == [["foo", 123], ["bar", 456], ["baz", 789]]

def test_job_result_format_failure():
    client = mock.MagicMock()
    job = model.Job(client, "12345", "presto", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=False)
    with pytest.raises(ValueError) as error:
        for row in job.result_format("msgpack.gz"):
            pass

def test_job_finished():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert not job(client, "queued").finished()
    assert not job(client, "booting").finished()
    assert not job(client, "running").finished()
    assert job(client, "success").finished()
    assert job(client, "error").finished()
    assert job(client, "killed").finished()

def test_job_success():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert not job(client, "queued").success()
    assert not job(client, "booting").success()
    assert not job(client, "running").success()
    assert job(client, "success").success()
    assert not job(client, "error").success()
    assert not job(client, "killed").success()

def test_job_error():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert not job(client, "queued").error()
    assert not job(client, "booting").error()
    assert not job(client, "running").error()
    assert not job(client, "success").error()
    assert job(client, "error").error()
    assert not job(client, "killed").error()

def test_job_killed():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert not job(client, "queued").killed()
    assert not job(client, "booting").killed()
    assert not job(client, "running").killed()
    assert not job(client, "success").killed()
    assert not job(client, "error").killed()
    assert job(client, "killed").killed()

def test_job_queued():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert job(client, "queued").queued()
    assert not job(client, "booting").queued()
    assert not job(client, "running").queued()
    assert not job(client, "success").queued()
    assert not job(client, "error").queued()
    assert not job(client, "killed").queued()

def test_job_running():
    def job(client, status):
        stub = model.Job(client, "1", "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        stub._update_progress = mock.MagicMock()
        return stub
    client = mock.MagicMock()
    client.job_status(return_value="testing")

    assert not job(client, "queued").running()
    assert not job(client, "booting").running()
    assert job(client, "running").running()
    assert not job(client, "success").running()
    assert not job(client, "error").running()
    assert not job(client, "killed").running()

def test_job_update_progress():
    def run(client, job_id, status):
        job = model.Job(client, job_id, "hive", "SELECT COUNT(1) FROM nasdaq", status=status)
        client.job_status.reset_mock()
        job._update_progress()

    client = mock.MagicMock()
    client.job_status(return_value="testing")

    run(client, "1", "queued")
    client.job_status.assert_called_with("1")

    run(client, "2", "booting")
    client.job_status.assert_called_with("2")

    run(client, "3", "running")
    client.job_status.assert_called_with("3")

    run(client, "4", "success")
    assert not client.job_status.called

    run(client, "5", "error")
    assert not client.job_status.called

    run(client, "6", "killed")
    assert not client.job_status.called

def test_job_update_status():
    client = mock.MagicMock()
    client.api.show_job = mock.MagicMock(return_value={
        "job_id": "67890",
        "type": "hive",
        "url": "http://console.example.com/jobs/67890",
        "query": "SELECT COUNT(1) FROM nasdaq",
        "status": "success",
        "debug": None,
        "start_at": datetime.datetime(2015, 2, 10, 0, 2, 14, tzinfo=dateutil.tz.tzutc()),
        "end_at": datetime.datetime(2015, 2, 10, 0, 2, 27, tzinfo=dateutil.tz.tzutc()),
        "cpu_time": None,
        "result_size": 22,
        "result": None,
        "result_url": None,
        "hive_result_schema": [["cnt", "bigint"]],
        "priority": 1,
        "retry_limit": 0,
        "org_name": None,
        "database": "sample_datasets",
    })
    job = model.Job(client, "67890", "hive", "SELECT COUNT(1) FROM nasdaq")
    job.finished = mock.MagicMock(return_value=False)
    assert job.status() == "success"
    client.api.show_job.assert_called_with("67890")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class Result(Model):
    """Result on Treasure Data Service
    """

    def __init__(self, client, name, url, org_name):
        super(Result, self).__init__(client)
        self._name = name
        self._url = url
        self._org_name = org_name

    @property
    def name(self):
        """
        TODO: add docstring
        """
        return self._name

    @property
    def url(self):
        """
        TODO: add docstring
        """
        return self._url

    @property
    def org_name(self):
        """
        TODO: add docstring
        """
        return self._org_name



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_result():
    client = mock.MagicMock()
    result = model.Result(client, "name", "url", "org_name")
    assert result.name == "name"
    assert result.url == "url"
    assert result.org_name == "org_name"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

class AccessControlAPI(object):
    ####
    ## Access Control API
    ##
    def grant_access_control(self, subject, action, scope, grant_option):
        """
        TODO: add docstring
        """
        params = {"subject": subject, "action": action, "scope": scope, "grant_option": str(grant_option)}
        with self.post("/v3/acl/grant", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Granting access control failed", res, body)
            return True

    def revoke_access_control(self, subject, action, scope):
        """
        TODO: add docstring
        """
        params = {"subject": subject, "action": action, "scope": scope}
        with self.post("/v3/acl/revoke", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Revoking access control failed", res, body)
            return True

    def test_access_control(self, user, action, scope):
        """
        TODO: add docstring
        [True, [{subject:str,action:str,scope:str}]]
        """
        params = {"user": user, "action": action, "scope": scope}
        with self.get("/v3/acl/test", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Testing access control failed", res, body)
            js = self.checked_json(body, ["permission", "access_controls"])
            perm = js["permission"]
            acl = [ [roleinfo["subject"], roleinfo["action"], roleinfo["scope"]] for roleinfo in js["access_controls"] ]
            return (perm, acl)

    def list_access_controls(self):
        """
        TODO: add docstring
        [{subject:str,action:str,scope:str}]
        """
        with self.get("/v3/acl/list") as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Listing access control failed", res, body)
            js = self.checked_json(body, ["access_controls"])
            acl = [ [roleinfo["subject"], roleinfo["action"], roleinfo["scope"], roleinfo["grant_option"]] for roleinfo in js["access_controls"] ]
            return acl



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import json
try:
    from urllib.parse import quote as urlquote # >=3.0
except ImportError:
    from urllib import quote as urlquote

class TableAPI(object):
    ####
    ## Table API
    ##

    def list_tables(self, db):
        """
        TODO: add docstring
        => {name:str => [type:str, count:int]}
        """
        with self.get("/v3/table/list/%s" % (urlquote(str(db)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List tables failed", res, body)
            js = self.checked_json(body, ["tables"])
            result = {}
            for m in js["tables"]:
                name = m.get("name")
                type = m.get("type", "?")
                count = int(m.get("count", 0))
                created_at = self.parsedate(self.get_or_else(m, "created_at", "1970-01-01T00:00:00Z"))
                updated_at = self.parsedate(self.get_or_else(m, "updated_at", "1970-01-01T00:00:00Z"))
                last_import = self.parsedate(self.get_or_else(m, "counter_updated_at", "1970-01-01T00:00:00Z"))
                last_log_timestamp = self.parsedate(self.get_or_else(m, "last_log_timestamp", "1970-01-01T00:00:00Z"))
                estimated_storage_size = int(m.get("estimated_storage_size", 0))
                schema = json.loads(m.get("schema", "[]"))
                expire_days = m.get("expire_days")
                primary_key = m.get("primary_key")
                primary_key_type = m.get("primary_key_type")
                result[name] = (type, schema, count, created_at, updated_at, estimated_storage_size, last_import, last_log_timestamp, expire_days, primary_key, primary_key_type)
            return result

    def create_log_table(self, db, table):
        """
        TODO: add docstring
        => True
        """
        return self._create_table(db, table, "log")

    def create_item_table(self, db, table, primary_key, primary_key_type):
        """
        TODO: add docstring
        => True
        """
        params = {"primary_key": primary_key, "primary_key_type": primary_key_type}
        return self._create_table(db, table, "item", params)

    def _create_table(self, db, table, type, params={}):
        with self.post("/v3/table/create/%s/%s/%s" % (urlquote(str(db)), urlquote(str(table)), urlquote(str(type))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Create %s table failed" % (type), res, body)
            return True

    def swap_table(self, db, table1, table2):
        """
        TODO: add docstring
        => True
        """
        with self.post("/v3/table/swap/%s/%s/%s" % (urlquote(str(db)), urlquote(str(table1)), urlquote(str(table2)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Swap tables failed", res, body)
            return True

    def update_schema(self, db, table, schema_json):
        """
        TODO: add docstring
        => True
        """
        with self.post("/v3/table/update-schema/%s/%s" % (urlquote(str(db)), urlquote(str(table))), {"schema": schema_json}) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Create schema table failed", res, body)
            return True

    def update_expire(self, db, table, expire_days):
        """
        TODO: add docstring
        """
        with self.post("/v3/table/update/%s/%s" % (urlquote(str(db)), urlquote(str(table))), {"expire_days": expire_days}) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Update table expiration failed", res, body)
            return True

    def delete_table(self, db, table):
        """
        TODO: add docstring
        => type:str
        """
        with self.post("/v3/table/delete/%s/%s" % (urlquote(str(db)), urlquote(str(table)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Delete table failed", res, body)
            js = self.checked_json(body, [])
            type = js.get("type", "?")
            return type



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_list_result_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "results": [
                {
                    "name": "foo",
                    "organization": null,
                    "url": "mysql://example.com/db/foo"
                },
                {
                    "name": "bar",
                    "organization": null,
                    "url": "postgresql://example.com/db/bar"
                },
                {
                    "name": "baz",
                    "organization": null,
                    "url": "s3://s3.example.com/baz.csv"
                }
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    results = td.list_result()
    td.get.assert_called_with("/v3/result/list")
    assert len(results) == 3
    assert results[0] == ("foo", "mysql://example.com/db/foo", None)
    assert results[1] == ("bar", "postgresql://example.com/db/bar", None)
    assert results[2] == ("baz", "s3://s3.example.com/baz.csv", None)

def test_list_result_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.list_result()
    assert error.value.args == ("500: List result table failed: error",)

def test_create_result_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    results = td.create_result("foo", "http://example.com")
    td.post.assert_called_with("/v3/result/create/foo", {"url": "http://example.com"})

def test_delete_result_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    results = td.delete_result("foo")
    td.post.assert_called_with("/v3/result/delete/foo")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import functools

from tdclient import pseudo_certifi as certifi

def exists(platform, path):
    if platform == "debian":
        if path == "/etc/ssl/certs/ca-certificates.crt":
            return True
    elif platform == "redhat":
        if path == "/etc/ssl/certs/ca-bundle.crt":
            return True
    return False

def test_certifi_debian():
    with mock.patch("os.path.exists") as os_path_exists:
        os_path_exists.side_effect = functools.partial(exists, "debian")
        assert certifi.where() == "/etc/ssl/certs/ca-certificates.crt"

def test_certifi_redhat():
    with mock.patch("os.path.exists") as os_path_exists:
        os_path_exists.side_effect = functools.partial(exists, "redhat")
        assert certifi.where() == "/etc/ssl/certs/ca-bundle.crt"

def test_certifi_unknown():
    with mock.patch("os.path.exists") as os_path_exists:
        os_path_exists.side_effect = functools.partial(exists, "unknown")
        assert certifi.where() is None



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import io
import msgpack
try:
    from unittest import mock
except ImportError:
    import mock
import pytest
import zlib

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_create_bulk_import_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.create_bulk_import("name", "db", "table")
    td.post.assert_called_with("/v3/bulk_import/create/name/db/table", {})

def test_delete_bulk_import_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.delete_bulk_import("name")
    td.post.assert_called_with("/v3/bulk_import/delete/name", {})

def test_show_bulk_import_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "status": "SUCCESS"
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    bulk_import = td.show_bulk_import("name")
    td.get.assert_called_with("/v3/bulk_import/show/name")
    assert bulk_import == {"status": "SUCCESS"}

def test_list_bulk_imports_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "bulk_imports":[
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    bulk_imports = td.list_bulk_imports()
    td.get.assert_called_with("/v3/bulk_import/list", {})
    assert len(bulk_imports) == 0

def test_list_bulk_imports_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.list_bulk_imports()
    assert error.value.args == ("500: List bulk imports failed: error",)

def test_list_bulk_import_parts_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "parts":[
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    parts = td.list_bulk_import_parts("foo")
    td.get.assert_called_with("/v3/bulk_import/list_parts/foo", {})
    assert len(parts) == 0

def test_list_bulk_import_upload_part_success():
    td = api.API("APIKEY")
    td.put = mock.MagicMock(return_value=make_response(200, b""))
    td.bulk_import_upload_part("name", "part_name", "stream", 1024)
    td.put.assert_called_with("/v3/bulk_import/upload_part/name/part_name", "stream", 1024)

def test_list_bulk_import_delete_part_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.bulk_import_delete_part("name", "part_name")
    td.post.assert_called_with("/v3/bulk_import/delete_part/name/part_name", {})

def test_freeze_bulk_import_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.freeze_bulk_import("name")
    td.post.assert_called_with("/v3/bulk_import/freeze/name", {})

def test_unfreeze_bulk_import_success():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.unfreeze_bulk_import("name")
    td.post.assert_called_with("/v3/bulk_import/unfreeze/name", {})

def test_perform_bulk_import_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "job_id": "12345"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    job_id = td.perform_bulk_import("name")
    td.post.assert_called_with("/v3/bulk_import/perform/name", {})
    assert job_id == "12345"

def test_commit_bulk_import_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    td.commit_bulk_import("name")
    td.post.assert_called_with("/v3/bulk_import/commit/name", {})

def msgpackb(list):
    """list -> bytes"""
    stream = io.BytesIO()
    packer = msgpack.Packer()
    for item in list:
        stream.write(packer.pack(item))
    return stream.getvalue()

def gzipb(bytes):
    """bytes -> bytes"""
    compress = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
    return compress.compress(bytes) + compress.flush()

def test_bulk_import_error_records_success():
    td = api.API("APIKEY")
    data = [
        {"str": "value1", "int": 1, "float": 2.3},
        {"str": "value4", "int": 5, "float": 6.7},
    ]
    body = gzipb(msgpackb(data))
    td.get = mock.MagicMock(return_value=make_response(200, body))
    rows = []
    for row in td.bulk_import_error_records("name"):
        rows.append(row)
    td.get.assert_called_with("/v3/bulk_import/error_records/name", {})

def test_bulk_import_error_records_false():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        for row in td.bulk_import_error_records("name"):
            pass
    td.get.assert_called_with("/v3/bulk_import/error_records/name", {})



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_access_control():
    client = mock.MagicMock()
    access_control = model.AccessControl(client, "subject", "action", "scope", "grant_option")
    assert access_control.subject == "subject"
    assert access_control.action == "action"
    assert access_control.scope == "scope"
    assert access_control.grant_option == "grant_option"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import gzip
import msgpack
import io
try:
    from urllib.parse import quote as urlquote # >=3.0
except ImportError:
    from urllib import quote as urlquote

class BulkImportAPI(object):
    ####
    ## Bulk import API
    ##

    def create_bulk_import(self, name, db, table, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/create/%s/%s/%s" % (urlquote(str(name)), urlquote(str(db)), urlquote(str(table))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Create bulk import failed", res, body)

    def delete_bulk_import(self, name, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/delete/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Delete bulk import failed", res, body)

    def show_bulk_import(self, name):
        """
        TODO: add docstring
        => data:dict
        """
        with self.get("/v3/bulk_import/show/%s" % (urlquote(str(name)))) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Show bulk import failed", res, body)
            js = self.checked_json(body, ["status"])
            return js

    def list_bulk_imports(self, params={}):
        """
        TODO: add docstring
        => result:[data:dict]
        """
        with self.get("/v3/bulk_import/list", params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List bulk imports failed", res, body)
            js = self.checked_json(body, ["bulk_imports"])
            return js["bulk_imports"]

    def list_bulk_import_parts(self, name, params={}):
        """
        TODO: add docstring
        """
        with self.get("/v3/bulk_import/list_parts/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("List bulk import parts failed", res, body)
            js = self.checked_json(body, ["parts"])
            return js["parts"]

    def bulk_import_upload_part(self, name, part_name, stream, size):
        """
        TODO: add docstring
        => None
        """
        with self.put("/v3/bulk_import/upload_part/%s/%s" % (urlquote(str(name)), urlquote(str(part_name))), stream, size) as res:
            code, body = res.status, res.read()
            if code / 100 != 2:
                self.raise_error("Upload a part failed", res, body)

    def bulk_import_delete_part(self, name, part_name, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/delete_part/%s/%s" % (urlquote(str(name)), urlquote(str(part_name))), params) as res:
            code, body = res.status, res.read()
            if code / 100 != 2:
                self.raise_error("Delete a part failed", res, body)

    def freeze_bulk_import(self, name, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/freeze/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Freeze bulk import failed", res, body)

    def unfreeze_bulk_import(self, name, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/unfreeze/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Unfreeze bulk import failed", res, body)

    def perform_bulk_import(self, name, params={}):
        """
        TODO: add docstring
        => jobId:str
        """
        with self.post("/v3/bulk_import/perform/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Perform bulk import failed", res, body)
            js = self.checked_json(body, ["job_id"])
            return str(js["job_id"])

    def commit_bulk_import(self, name, params={}):
        """
        TODO: add docstring
        => None
        """
        with self.post("/v3/bulk_import/commit/%s" % (urlquote(str(name))), params) as res:
            code, body = res.status, res.read()
            if code != 200:
                self.raise_error("Commit bulk import failed", res, body)

    def bulk_import_error_records(self, name, params={}):
        """
        TODO: add docstring
        => data...
        """
        with self.get("/v3/bulk_import/error_records/%s" % (urlquote(str(name))), params) as res:
            code = res.status
            if code != 200:
                body = res.read()
                self.raise_error("Failed to get bulk import error records", res, body)

            body = io.BytesIO(res.read())
            decompressor = gzip.GzipFile(fileobj=body)

            content_type = res.getheader("content-type", "application/x-msgpack; charset=utf-8")
            type_params = dict([ p.strip().split("=", 2) for p in content_type.split(";") if 0 < p.find("=") ])

            unpacker = msgpack.Unpacker(decompressor, encoding=str(type_params.get("charset", "utf-8")))
            for row in unpacker:
                yield row



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_server_status_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "status": "ok"
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    assert td.server_status() == "ok"
    td.get.assert_called_with("/v3/system/server_status")

def test_server_status_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    assert td.server_status() == "Server is down (500)"
    td.get.assert_called_with("/v3/system/server_status")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import datetime
import dateutil.tz
import json
try:
    from unittest import mock
except ImportError:
    import mock
import msgpack
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_list_jobs_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "count":11,"from":0,"to":10,"jobs":[
                {"status":"success","cpu_time":496570,"result_size":24,"duration":262,"job_id":"18882028","created_at":"2014-12-23 12:00:18 UTC","updated_at":"2014-12-23 12:04:42 UTC","start_at":"2014-12-23 12:00:19 UTC","end_at":"2014-12-23 12:04:41 UTC","query":"select count(1) from import_test","type":"hive","priority":0,"retry_limit":0,"result":"","url":"http://console.treasuredata.com/jobs/18882028","user_name":"owner","hive_result_schema":"[[\\"_c0\\", \\"bigint\\"]]","organization":null,"database":"jma_weather"},
                {"status":"success","cpu_time":489540,"result_size":24,"duration":272,"job_id":"18880612","created_at":"2014-12-23 11:00:16 UTC","updated_at":"2014-12-23 11:04:48 UTC","start_at":"2014-12-23 11:00:16 UTC","end_at":"2014-12-23 11:04:48 UTC","query":"select count(1) from import_test","type":"hive","priority":0,"retry_limit":0,"result":"","url":"http://console.treasuredata.com/jobs/18880612","user_name":"owner","hive_result_schema":"[[\\"_c0\\", \\"bigint\\"]]","organization":null,"database":"jma_weather"},
                {"status":"success","cpu_time":486630,"result_size":24,"duration":263,"job_id":"18879199","created_at":"2014-12-23 10:00:21 UTC","updated_at":"2014-12-23 10:04:44 UTC","start_at":"2014-12-23 10:00:21 UTC","end_at":"2014-12-23 10:04:44 UTC","query":"select count(1) from import_test","type":"hive","priority":0,"retry_limit":0,"result":"","url":"http://console.treasuredata.com/jobs/18879199","user_name":"owner","hive_result_schema":"[[\\"_c0\\", \\"bigint\\"]]","organization":null,"database":"jma_weather"}
            ]
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    jobs = td.list_jobs(0, 2)
    td.get.assert_called_with("/v3/job/list", {"from": "0", "to": "2"})
    assert len(jobs) == 3
    assert sorted([ job["job_id"] for job in jobs ]) == ["18879199", "18880612", "18882028"]
    assert sorted([ job["url"] for job in jobs ]) == [
        "http://console.treasuredata.com/jobs/18879199",
        "http://console.treasuredata.com/jobs/18880612",
        "http://console.treasuredata.com/jobs/18882028",
    ]
    assert [ job["debug"] for job in jobs ] == [None, None, None]
    assert sorted([ job["hive_result_schema"] for job in jobs ]) == [
        [["_c0", "bigint"]],
        [["_c0", "bigint"]],
        [["_c0", "bigint"]],
    ]

def test_list_jobs_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.list_jobs(0, 2)
    assert error.value.args == ("500: List jobs failed: error",)

def test_show_job_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "cpu_time": null,
            "created_at": "2015-02-09 11:44:25 UTC",
            "database": "sample_datasets",
            "debug": {
                "cmdout": "started at 2015-02-09T11:44:27Z\\nexecuting query: SELECT COUNT(1) FROM nasdaq\\n",
                "stderr": null
            },
            "duration": 1,
            "end_at": "2015-02-09 11:44:28 UTC",
            "hive_result_schema": "[[\\"cnt\\", \\"bigint\\"]]",
            "job_id": "12345",
            "organization": null,
            "priority": 1,
            "query": "SELECT COUNT(1) FROM nasdaq",
            "result": "",
            "result_size": 22,
            "retry_limit": 0,
            "start_at": "2015-02-09 11:44:27 UTC",
            "status": "success",
            "type": "presto",
            "updated_at": "2015-02-09 11:44:28 UTC",
            "url": "http://console.example.com/jobs/12345",
            "user_name": "nobody@example.com"
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    job = td.show_job(12345)
    td.get.assert_called_with("/v3/job/show/12345")
    assert job["job_id"] == 12345
    assert job["type"] == "presto"
    assert job["url"] == "http://console.example.com/jobs/12345"
    assert job["query"] == "SELECT COUNT(1) FROM nasdaq"
    assert job["status"] == "success"
    assert job["debug"] == {
        "cmdout": "started at 2015-02-09T11:44:27Z\nexecuting query: SELECT COUNT(1) FROM nasdaq\n",
        "stderr": None,
    }
    assert job["start_at"] == datetime.datetime(2015, 2, 9, 11, 44, 27, tzinfo=dateutil.tz.tzutc())
    assert job["end_at"] == datetime.datetime(2015, 2, 9, 11, 44, 28, tzinfo=dateutil.tz.tzutc())
    assert job["cpu_time"] is None
    assert job["result_size"] == 22
    assert job["result"] is None
    assert job["result_url"] is None
    assert job["hive_result_schema"] == [["cnt", "bigint"]]
    assert job["priority"] == 1
    assert job["retry_limit"] == 0
    assert job["org_name"] is None
    assert job["database"] == "sample_datasets"

def test_job_status_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "status": "RUNNING"
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    jobs = td.job_status(12345)
    td.get.assert_called_with("/v3/job/status/12345")

def test_job_result_success():
    td = api.API("APIKEY")
    packer = msgpack.Packer()
    rows = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    body = b""
    for row in rows:
        body += packer.pack(row)
    td.get = mock.MagicMock(return_value=make_response(200, body))
    result = td.job_result(12345)
    td.get.assert_called_with("/v3/job/result/12345", {"format": "msgpack"})
    assert result == rows

def test_job_result_each_success():
    td = api.API("APIKEY")
    packer = msgpack.Packer()
    rows = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    body = b""
    for row in rows:
        body += packer.pack(row)
    td.get = mock.MagicMock(return_value=make_response(200, body))
    result = []
    for row in td.job_result_each(12345):
        result.append(row)
    td.get.assert_called_with("/v3/job/result/12345", {"format": "msgpack"})
    assert result == rows

def test_job_result_json_success():
    td = api.API("APIKEY")
    rows = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    # result will be a JSON record per line (#4)
    body = "\n".join([ json.dumps(row) for row in rows ]).encode("utf-8")
    td.get = mock.MagicMock(return_value=make_response(200, body))
    result = td.job_result_format(12345, "json")
    td.get.assert_called_with("/v3/job/result/12345", {"format": "json"})
    assert result == rows

def test_job_result_json_each_success():
    td = api.API("APIKEY")
    rows = [
        ["foo", 123],
        ["bar", 456],
        ["baz", 789],
    ]
    # result will be a JSON record per line (#4)
    body = "\n".join([ json.dumps(row) for row in rows ]).encode("utf-8")
    td.get = mock.MagicMock(return_value=make_response(200, body))
    result = []
    for row in td.job_result_format_each(12345, "json"):
        result.append(row)
    td.get.assert_called_with("/v3/job/result/12345", {"format": "json"})
    assert result == rows

def test_kill_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "former_status": "foo"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    jobs = td.kill(12345)
    td.post.assert_called_with("/v3/job/kill/12345")

def test_hive_query_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "job": "12345",
            "database": "sample_datasets",
            "job_id": "12345"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    job_id = td.hive_query("SELECT COUNT(1) FROM nasdaq", db="sample_datasets")
    td.post.assert_called_with("/v3/job/issue/hive/sample_datasets", {"query": "SELECT COUNT(1) FROM nasdaq"})
    assert job_id == "12345"

def test_pig_query_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "job": "12345",
            "database": "sample_datasets",
            "job_id": "12345"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    job_id = td.pig_query("A=LOAD 'nasdaq';B=GROUP A ALL;C=FOREACH B GENERATE COUNT(A);", db="sample_datasets")
    td.post.assert_called_with("/v3/job/issue/pig/sample_datasets", {"query": "A=LOAD 'nasdaq';B=GROUP A ALL;C=FOREACH B GENERATE COUNT(A);"})
    assert job_id == "12345"

def test_presto_query_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "job": "12345",
            "database": "sample_datasets",
            "job_id": "12345"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    job_id = td.query("SELECT COUNT(1) FROM nasdaq", db="sample_datasets", type="presto", priority=0)
    td.post.assert_called_with("/v3/job/issue/presto/sample_datasets", {"query": "SELECT COUNT(1) FROM nasdaq", "priority": 0})
    assert job_id == "12345"

def test_query_success():
    td = api.API("APIKEY")
    body = b"""
        {
            "job": "12345",
            "database": "sample_datasets",
            "job_id": "12345"
        }
    """
    td.post = mock.MagicMock(return_value=make_response(200, body))
    job_id = td.query("SELECT COUNT(1) FROM nasdaq", db="sample_datasets", priority="HIGH")
    td.post.assert_called_with("/v3/job/issue/hive/sample_datasets", {"query": "SELECT COUNT(1) FROM nasdaq", "priority": 1})
    assert job_id == "12345"

def test_query_priority_unknown():
    td = api.API("APIKEY")
    td.post = mock.MagicMock(return_value=make_response(200, b""))
    with pytest.raises(ValueError) as error:
        td.query("SELECT COUNT(1) FROM nasdaq", db="sample_datasets", priority="unknown")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class BulkImport(Model):
    """Bulk-import session on Treasure Data Service
    """

    def __init__(self, client, name=None, database=None, table=None, status=None, upload_frozen=None, job_id=None, valid_records=None, error_records=None, valid_parts=None, error_parts=None, **kwargs):
        super(BulkImport, self).__init__(client)
        self._name = name
        self._database = database
        self._table = table
        self._status = status
        self._upload_frozen = upload_frozen
        self._job_id = job_id
        self._valid_records = valid_records
        self._error_records = error_records
        self._valid_parts = valid_parts
        self._error_parts = error_parts

    @property
    def name(self):
        """
        Returns: name of the bulk import session
        """
        return self._name

    @property
    def database(self):
        """
        Returns: database name in a string which the bulk import session is working on
        """
        return self._database

    @property
    def table(self):
        """
        Returns: table name in a string which the bulk import session is working on
        """
        return self._table

    @property
    def status(self):
        """
        Returns: status of the bulk import session in a string
        """
        return self._status

    @property
    def job_id(self):
        """
        TODO: add docstring
        """
        return self._job_id

    @property
    def valid_records(self):
        """
        TODO: add docstring
        """
        return self._valid_records

    @property
    def error_records(self):
        """
        TODO: add docstring
        """
        return self._error_records

    @property
    def valid_parts(self):
        """
        TODO: add docstring
        """
        return self._valid_parts

    @property
    def error_parts(self):
        """
        TODO: add docstring
        """
        return self._error_parts

    @property
    def upload_frozen(self):
        """
        TODO: add docstring
        """
        return self._upload_frozen

    def error_record_items(self):
        """
        TODO: add docstring
        """
        for record in self._client.bulk_import_error_records(self.name):
            yield record



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_table():
    client = mock.MagicMock()
    table = model.Table(client, "db_name", "table_name", "type", "schema", 12345, created_at="created_at", updated_at="updated_at", estimated_storage_size=67890, last_import="last_import", last_log_timestamp="last_log_timestamp", expire_days="expire_days", primary_key="primary_key", primary_key_type="primary_key_type")
    assert table.type == "type"
    assert table.db_name == "db_name"
    assert table.table_name == "table_name"
    assert table.schema == "schema"
    assert table.count == 12345
    assert table.estimated_storage_size == 67890
    assert table.primary_key == "primary_key"
    assert table.primary_key_type == "primary_key_type"
    assert table.database_name == "db_name"
    assert table.name == "table_name"
    assert table.created_at == "created_at"
    assert table.updated_at == "updated_at"
    assert table.last_import == "last_import"
    assert table.last_log_timestamp == "last_log_timestamp"
    assert table.expire_days == "expire_days"
    assert table.identifier == "db_name.table_name"

def test_table_permission():
    client = mock.MagicMock()
    table = model.Table(client, "sample_datasets", "nasdaq", "log", "schema", 12345)
    client.database().permission = "permission"
    assert table.permission == "permission"
    client.database.assert_called_with("sample_datasets")

def test_table_estimated_storage_size_string():
    client = mock.MagicMock()
    table1 = model.Table(client, "db_name", "table_name", "type", "schema", 12345, estimated_storage_size=1)
    assert table1.estimated_storage_size_string == "0.0 GB"
    table2 = model.Table(client, "db_name", "table_name", "type", "schema", 12345, estimated_storage_size=50*1024*1024)
    assert table2.estimated_storage_size_string == "0.01 GB"
    table3 = model.Table(client, "db_name", "table_name", "type", "schema", 12345, estimated_storage_size=50*1024*1024*1024)
    assert table3.estimated_storage_size_string == "50.0 GB"
    table4 = model.Table(client, "db_name", "table_name", "type", "schema", 12345, estimated_storage_size=300*1024*1024*1024)
    assert table4.estimated_storage_size_string == "300 GB"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from array import array
try:
    import certifi
except ImportError:
    # on py27 on windows, it tries to import pseudo "certifi" module instead of original one.
    # to avoid the weird behavior, the pseudo "certifi" module should be named differently.
    from tdclient import pseudo_certifi as certifi
import contextlib
import dateutil.parser
import email.utils
import io
import json
import logging
import os
import socket
import ssl
import sys
import time
try:
    import urllib.parse as urlparse # >=3.0
except ImportError:
    import urlparse
import urllib3
import zlib

from tdclient.access_control_api import AccessControlAPI
from tdclient.account_api import AccountAPI
from tdclient.bulk_import_api import BulkImportAPI
from tdclient.database_api import DatabaseAPI
from tdclient.export_api import ExportAPI
from tdclient.import_api import ImportAPI
from tdclient.job_api import JobAPI
from tdclient.partial_delete_api import PartialDeleteAPI
from tdclient.result_api import ResultAPI
from tdclient.schedule_api import ScheduleAPI
from tdclient.server_status_api import ServerStatusAPI
from tdclient.table_api import TableAPI
from tdclient.user_api import UserAPI
from tdclient import version

log = logging.getLogger(__name__)

class ParameterValidationError(Exception):
    pass

# Generic API error
class APIError(Exception):
    pass

# 401 API errors
class AuthError(APIError):
    pass

# 403 API errors, used for database permissions
class ForbiddenError(APIError):
    pass

# 409 API errors
class AlreadyExistsError(APIError):
    pass

# 404 API errors
class NotFoundError(APIError):
    pass

class API(AccessControlAPI, AccountAPI, BulkImportAPI, DatabaseAPI, ExportAPI, ImportAPI,
          JobAPI, PartialDeleteAPI, ResultAPI, ScheduleAPI, ServerStatusAPI, TableAPI, UserAPI):
    """Internal API class

    Args:
        apikey (str): the API key of Treasure Data Service. If `None` is given, `TD_API_KEY` will be used if available.
        user_agent (str): custom User-Agent.
        endpoint (str): custom endpoint URL. If `None` is given, `TD_API_SERVER` will be used if available.
        headers (dict): custom HTTP headers.
        retry_post_requests (bool): Specify whether allowing API client to retry POST requests. `False` by default.
        max_cumul_retry_delay (int): maximum retry limit in seconds. 600 seconds by default.
        http_proxy (str): HTTP proxy setting. if `None` is given, `HTTP_PROXY` will be used if available.
    """

    DEFAULT_ENDPOINT = "https://api.treasuredata.com/"
    DEFAULT_IMPORT_ENDPOINT = "https://api-import.treasuredata.com/"

    def __init__(self, apikey=None, user_agent=None, endpoint=None, headers={}, retry_post_requests=False,
                 max_cumul_retry_delay=600, http_proxy=None, **kwargs):
        if apikey is not None:
            self._apikey = apikey
        elif "TD_API_KEY" in os.environ:
            self._apikey = os.getenv("TD_API_KEY")
        else:
            raise(ValueError("no API key given"))

        if user_agent is not None:
            self._user_agent = user_agent
        else:
            self._user_agent = "TD-Client-Python: %s" % (version.__version__)

        if endpoint is not None:
            self._endpoint = endpoint
        elif os.getenv("TD_API_SERVER"):
            self._endpoint = os.getenv("TD_API_SERVER")
        else:
            self._endpoint = self.DEFAULT_ENDPOINT

        pool_options = dict(kwargs)
        certs = pool_options.get("ca_certs", certifi.where())
        if certs is not None:
            pool_options["ca_certs"] = certs
            pool_options["cert_reqs"] = ssl.CERT_REQUIRED

        if "connect_timeout" in pool_options or "read_timeout" in pool_options or "send_timeout" in pool_options:
            connect_timeout = pool_options.pop("connect_timeout") if "connect_timeout" in pool_options else 0
            read_timeout = pool_options.pop("read_timeout") if "read_timeout" in pool_options else 0
            send_timeout = pool_options.pop("send_timeout") if "send_timeout" in pool_options else 0
            pool_options["timeout"] = connect_timeout + read_timeout + send_timeout

        if http_proxy is None and "HTTP_PROXY" in os.environ:
            http_proxy = os.getenv("HTTP_PROXY")

        if http_proxy is None:
            self.http = urllib3.PoolManager(**pool_options)
        else:
            if http_proxy.startswith("http://"):
                self.http = urllib3.ProxyManager(http_proxy, **pool_options)
            else:
                self.http = urllib3.ProxyManager("http://%s" % http_proxy, **pool_options)

        self._retry_post_requests = retry_post_requests
        self._max_cumul_retry_delay = max_cumul_retry_delay
        self._headers = dict([ (key.lower(), value) for (key, value) in headers.items() ])

    @property
    def apikey(self):
        return self._apikey

    @property
    def endpoint(self):
        return self._endpoint

    def get(self, path, params={}, **kwargs):
        headers = {
            "accept-encoding": "deflate, gzip",
        }
        url, headers = self.build_request(path=path, headers=headers, **kwargs)

        log.debug("REST GET call:\n  headers: %s\n  path: %s\n  params: %s" % (repr(headers), repr(path), repr(params)))

        # up to 7 retries with exponential (base 2) back-off starting at 'retry_delay'
        retry_delay = 5
        cumul_retry_delay = 0

        # for both exceptions and 500+ errors retrying is enabled by default.
        # The total number of retries cumulatively should not exceed 10 minutes / 600 seconds
        response = None
        while True:
            try:
                response = self.http.request("GET", url, fields=params, headers=headers, decode_content=True, preload_content=False)
                # retry if the HTTP error code is 500 or higher and we did not run out of retrying attempts
                if response.status < 500:
                    break
                else:
                    log.warn("Error %d: %s. Retrying after %d seconds..." % (response.status, response.data, retry_delay))
            except ( urllib3.exceptions.TimeoutStateError, urllib3.exceptions.TimeoutError, urllib3.exceptions.PoolError, socket.error ):
                pass

            if cumul_retry_delay <= self._max_cumul_retry_delay:
                log.warn("Retrying after %d seconds..." % (retry_delay))
                self.sleep(retry_delay)
                cumul_retry_delay += retry_delay
                retry_delay *= 2
            else:
                raise(APIError("Retrying stopped after %d seconds." % (self._max_cumul_retry_delay)))

        log.debug("REST GET response:\n  headers: %s\n  status: %d\n  body: <omitted>" % (repr(dict(response.getheaders())), response.status))

        return contextlib.closing(response)

    def post(self, path, params={}, **kwargs):
        url, headers = self.build_request(path=path, headers={}, **kwargs)

        log.debug("REST POST call:\n  headers: %s\n  path: %s\n  params: %s" % (repr(headers), repr(path), repr(params)))

        # up to 7 retries with exponential (base 2) back-off starting at 'retry_delay'
        retry_delay = 5
        cumul_retry_delay = 0

        # for both exceptions and 500+ errors retrying can be enabled by initialization
        # parameter 'retry_post_requests'. The total number of retries cumulatively
        # should not exceed 10 minutes / 600 seconds
        response = None
        while True:
            try:
                response = self.http.request("POST", url, fields=params, headers=headers, decode_content=True, preload_content=False)
                # if the HTTP error code is 500 or higher and the user requested retrying
                # on post request, attempt a retry
                if response.status < 500:
                    break
                else:
                    if not self._retry_post_requests:
                        raise(APIError("Retrying stopped by retry_post_requests == False"))
                    log.warn("Error %d: %s. Retrying after %d seconds..." % (response.status, response.data, retry_delay))
            except ( urllib3.exceptions.TimeoutStateError, urllib3.exceptions.TimeoutError, urllib3.exceptions.PoolError, socket.error ):
                if self._retry_post_requests:
                    raise(APIError("Retrying stopped by retry_post_requests == False"))

            if cumul_retry_delay <= self._max_cumul_retry_delay:
                log.warn("Retrying after %d seconds..." % (retry_delay))
                self.sleep(retry_delay)
                cumul_retry_delay += retry_delay
                retry_delay *= 2
            else:
                raise(APIError("Retrying stopped after %d seconds." % (self._max_cumul_retry_delay)))

        log.debug("REST POST response:\n  headers: %s\n  status: %d\n  body: <omitted>" % (repr(dict(response.getheaders())), response.status))

        return contextlib.closing(response)

    def put(self, path, bytes_or_stream, size, **kwargs):
        headers = {}
        headers["content-length"] = str(size)
        headers["content-type"] = "application/octet-stream"
        url, headers = self.build_request(path=path, headers=headers, **kwargs)

        log.debug("REST PUT call:\n  headers: %s\n  path: %s\n  body: <omitted>" % (repr(headers), repr(path)))

        if hasattr(bytes_or_stream, "read"):
            # file-like must support `read` and `fileno` to work with `httplib`
            fileno_supported = hasattr(bytes_or_stream, "fileno")
            if fileno_supported:
                try:
                    bytes_or_stream.fileno()
                except io.UnsupportedOperation:
                    # `io.BytesIO` doesn't support `fileno`
                    fileno_supported = False
            if fileno_supported:
                stream = bytes_or_stream
            else:
                stream = array(str("b"), bytes_or_stream.read())
            
        else:
            # send request body as an `array.array` since `httplib` requires the request body to be a unicode string
            stream = array(str("b"), bytes_or_stream)

        # up to 7 retries with exponential (base 2) back-off starting at 'retry_delay'
        retry_delay = 5
        cumul_retry_delay = 0

        response = None
        while True:
            try:
                response = self.http.urlopen("PUT", url, body=stream, headers=headers, decode_content=True, preload_content=False)
                if response.status < 500:
                    break
                else:
                    log.warn("Error %d: %s. Retrying after %d seconds..." % (response.status, response.data, retry_delay))
            except ( urllib3.exceptions.TimeoutStateError, urllib3.exceptions.TimeoutError, urllib3.exceptions.PoolError, socket.error ):
                pass

            if cumul_retry_delay <= self._max_cumul_retry_delay:
                log.warn("Retrying after %d seconds..." % (retry_delay))
                self.sleep(retry_delay)
                cumul_retry_delay += retry_delay
                retry_delay *= 2
            else:
                raise(APIError("Retrying stopped after %d seconds." % (self._max_cumul_retry_delay)))

        log.debug("REST PUT response:\n  headers: %s\n  status: %d\n  body: <omitted>" % (repr(dict(response.getheaders())), response.status))

        return contextlib.closing(response)

    def build_request(self, path=None, headers={}, endpoint=None):
        if endpoint is None:
            endpoint = self._endpoint
        if path is None:
            url = endpoint
        else:
            p = urlparse.urlparse(endpoint)
            # should not use `os.path.join` since it returns path string like "/foo\\bar"
            request_path = path if p.path == "/" else "/".join([p.path, path])
            url = urlparse.urlunparse(urlparse.ParseResult(p.scheme, p.netloc, request_path, p.params, p.query, p.fragment))
        # use default headers first
        _headers = dict(self._headers)
        # add default headers
        _headers["authorization"] = "TD1 %s" % (self._apikey,)
        _headers["date"] = email.utils.formatdate(time.time())
        _headers["user-agent"] = self._user_agent
        # override given headers
        _headers.update(dict([ (key.lower(), value) for (key, value) in headers.items() ]))
        return (url, _headers)

    def raise_error(self, msg, res, body):
        status_code = res.status
        s = body.decode("utf-8")
        if status_code == 404:
            raise NotFoundError("%s: %s" % (msg, s))
        elif status_code == 409:
            raise AlreadyExistsError("%s: %s" % (msg, s))
        elif status_code == 401:
            raise AuthError("%s: %s" % (msg, s))
        elif status_code == 403:
            raise ForbiddenError("%s: %s" % (msg, s))
        else:
            raise APIError("%d: %s: %s" % (status_code, msg, s))

    def checked_json(self, body, required):
        js = None
        try:
            js = json.loads(body.decode("utf-8"))
        except ValueError as error:
            raise APIError("Unexpected API response: %s: %s" % (error, repr(body)))
        js = dict(js)
        if 0 < [ k in js for k in required ].count(False):
            missing = [ k for k in required if k not in js ]
            raise APIError("Unexpected API response: %s: %s" % (repr(missing), repr(body)))
        return js

    def sleep(self, secs):
        time.sleep(secs)

    def parsedate(self, s):
        # TODO: parse datetime with specific format
        # for now, this parses string without format since API may return ambiguous value :(
        return dateutil.parser.parse(s)

    def get_or_else(self, hashmap, key, default_value=None):
        value = hashmap.get(key)
        return default_value if value is None else value

    def close(self):
        # urllib3 doesn't allow to close all connections immediately.
        # all connections in pool will be closed eventually during gc.
        self.http.clear()



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_database():
    client = mock.MagicMock()
    database = model.Database(client, "sample_datasets", tables=["nasdaq", "www_access"], count=12345, created_at="created_at", updated_at="updated_at", org_name="org_name", permission="administrator")
    assert database.org_name == "org_name"
    assert database.permission == "administrator"
    assert database.count == 12345
    assert database.name == "sample_datasets"
    assert database.tables() == ["nasdaq", "www_access"]
    assert database.created_at == "created_at"
    assert database.updated_at == "updated_at"

def test_database_update_tables():
    client = mock.MagicMock()
    client.tables = mock.MagicMock(return_value=[
        model.Table(client, "sample_datasets", "foo", "type", "schema", "count"),
        model.Table(client, "sample_datasets", "bar", "type", "schema", "count"),
        model.Table(client, "sample_datasets", "baz", "type", "schema", "count"),
    ])
    database = model.Database(client, "sample_datasets", tables=None, count=12345, created_at="created_at", updated_at="updated_at", org_name="org_name", permission="administrator")
    tables = database.tables()
    assert [ table.name for table in tables ] == ["foo", "bar", "baz"]
    client.tables.assert_called_with("sample_datasets")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model
from tdclient.job_model import Job

class ScheduledJob(Job):
    """Scheduled job on Treasure Data Service
    """
    def __init__(self, client, scheduled_at, *args, **kwargs):
        super(ScheduledJob, self).__init__(client, *args, **kwargs)
        self._scheduled_at = scheduled_at

    @property
    def scheduled_at(self):
        """
        Returns: a :class:`datetime.datetime` represents the schedule of next invocation of the job
        """
        return self._created_at

class Schedule(Model):
    """Schedule on Treasure Data Service
    """

    def __init__(self, client, name, cron, query, database=None, result_url=None, timezone=None, delay=None, next_time=None, priority=None, retry_limit=None, org_name=None):
        super(Schedule, self).__init__(client)
        self._name = name
        self._cron = cron
        self._query = query
        self._database = database
        self._result_url = result_url
        self._timezone = timezone
        self._delay = delay
        self._next_time = next_time
        self._priority = priority
        self._retry_limit = retry_limit
        self._org_name = org_name

    @property
    def name(self):
        """
        TODO: add docstring
        """
        return self._name

    @property
    def cron(self):
        """
        TODO: add docstring
        """
        return self._cron

    @property
    def query(self):
        """
        TODO: add docstring
        """
        return self._query

    @property
    def database(self):
        """
        TODO: add docstring
        """
        return self._database

    @property
    def result_url(self):
        """
        TODO: add docstring
        """
        return self._result_url

    @property
    def timezone(self):
        """
        TODO: add docstring
        """
        return self._timezone

    @property
    def delay(self):
        """
        TODO: add docstring
        """
        return self._delay

    @property
    def priority(self):
        """
        TODO: add docstring
        """
        return self._priority

    @property
    def retry_limit(self):
        """
        TODO: add docstring
        """
        return self._retry_limit

    @property
    def org_name(self):
        """
        TODO: add docstring
        """
        return self._org_name

    @property
    def next_time(self):
        """
        TODO: add docstring
        """
        return self._next_time

    def run(self, time, num):
        """
        TODO: add docstring
        """
        return self._client.run_schedule(time, num)



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import contextlib
import gzip
import json
import msgpack
import os
import tempfile
try:
    import urllib.parse as urlparse # >=3.0
except ImportError:
    import urlparse
try:
    from urllib.parse import quote as urlquote # >=3.0
except ImportError:
    from urllib import quote as urlquote

class ImportAPI(object):
    ####
    ## Import API
    ##

    def import_data(self, db, table, format, bytes_or_stream, size, unique_id=None):
        """Import data into Treasure Data Service

        This method expects data from a file-like object formatted with "msgpack.gz".

        Params:
            db (str): name of a database
            table (str): name of a table
            format (str): format of data type (e.g. "msgpack.gz")
            bytes_or_stream (str or file-like): a byte string or a file-like object contains the data
            size (int): the length of the data
            unique_id (str): a unique identifier of the data

        Returns: float represents the elapsed time to import data
        """
        if unique_id is not None:
            path = "/v3/table/import_with_id/%s/%s/%s/%s" % (urlquote(str(db)), urlquote(str(table)), urlquote(str(unique_id)), urlquote(str(format)))
        else:
            path = "/v3/table/import/%s/%s/%s" % (urlquote(str(db)), urlquote(str(table)), urlquote(str(format)))

        kwargs = {}
        with self.put(path, bytes_or_stream, size, **kwargs) as res:
            code, body = res.status, res.read()
            if code / 100 != 2:
                self.raise_error("Import failed", res, body)
            js = self.checked_json(body, ["elapsed_time"])
            time = float(js["elapsed_time"])
            return time

    def import_file(self, db, table, format, file, unique_id=None):
        """Import data into Treasure Data Service, from an existing file on filesystem.

        This method will decompress/deserialize records from given file, and then
        convert it into format acceptable from Treasure Data Service ("msgpack.gz").

        Params:
            db (str): name of a database
            table (str): name of a table
            format (str): format of data type (e.g. "msgpack", "json")
            file (str or file-like): a name of a file, or a file-like object contains the data
            unique_id (str): a unique identifier of the data

        Returns: float represents the elapsed time to import data
        """
        if hasattr(file, "read"):
            if format.endswith(".gz"):
                return self._import_file(db, table, format[0:len(format)-len(".gz")], gzip.GzipFile(fileobj=file), unique_id=unique_id)
            else:
                return self._import_file(db, table, format, file, unique_id=unique_id)
        else:
            with open(file, "rb") as fp:
                if format.endswith(".gz"):
                    return self._import_file(db, table, format[0:len(format)-len(".gz")], gzip.GzipFile(fileobj=fp), unique_id=unique_id)
                else:
                    return self._import_file(db, table, format, fp, unique_id=unique_id)

    def _import_file(self, db, table, format, file, unique_id=None):
        if format == "msgpack":
            return self._import_items(db, table, self._parse_msgpack_file(file), unique_id=unique_id)
        elif format == "json":
            return self._import_items(db, table, self._parse_json_file(file), unique_id=unique_id)
        else:
            raise TypeError("unknown format: %s" % (format,))

    def _import_items(self, db, table, items, unique_id=None):
        with tempfile.TemporaryFile() as fp:
            with contextlib.closing(gzip.GzipFile(mode="wb", fileobj=fp)) as gz:
                packer = msgpack.Packer()
                for record in items:
                    gz.write(packer.pack(record))
            fp.seek(0)
            size = os.fstat(fp.fileno()).st_size
            return self.import_data(db, table, "msgpack.gz", fp, size, unique_id=unique_id)

    def _parse_msgpack_file(self, file):
        # current impl doesn't torelate any unpack error
        unpacker = msgpack.Unpacker(file)
        for record in unpacker:
            yield record

    def _parse_json_file(self, file):
        # current impl doesn't torelate any JSON parse error
        for s in file:
            record = json.loads(s.decode("utf-8"))
            yield record



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

class ServerStatusAPI(object):
    ####
    ## Server Status API
    ##

    def server_status(self):
        """
        TODO: add docstring
        => status:str
        """
        with self.get("/v3/system/server_status") as res:
            code, body = res.status, res.read()
            if code != 200:
                return "Server is down (%d)" % (code,)
            js = self.checked_json(body, ["status"])
            status = js["status"]
            return status



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_account():
    client = mock.MagicMock()
    account = model.Account(client, 1, 2, storage_size=3456, guaranteed_cores=7, maximum_cores=8, created_at="created_at")
    assert account.account_id == 1
    assert account.plan == 2
    assert account.storage_size == 3456
    assert account.guaranteed_cores == 7
    assert account.maximum_cores == 8
    assert account.created_at == "created_at"

def test_account_storage_size_string():
    client = mock.MagicMock()
    account1 = model.Account(client, 1, 1, storage_size=0)
    assert account1.storage_size_string == "0.0 GB"
    account2 = model.Account(client, 1, 1, storage_size=50*1024*1024)
    assert account2.storage_size_string == "0.01 GB"
    account3 = model.Account(client, 1, 1, storage_size=50*1024*1024*1024)
    assert account3.storage_size_string == "50.0 GB"
    account4 = model.Account(client, 1, 1, storage_size=300*1024*1024*1024)
    assert account4.storage_size_string == "300 GB"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient._model import Model

class AccessControl(Model):
    """Access control settings of a user on Treasure Data Service
    """

    def __init__(self, client, subject, action, scope, grant_option):
        super(AccessControl, self).__init__(client)
        self._subject = subject
        self._action = action
        self._scope = scope
        self._grant_option = grant_option

    @property
    def subject(self):
        """
        TODO: add docstring
        """
        return self._subject

    @property
    def action(self):
        """
        TODO: add docstring
        """
        return self._action

    @property
    def scope(self):
        """
        TODO: add docstring
        """
        return self._scope

    @property
    def grant_option(self):
        """
        TODO: add docstring
        """
        return self._grant_option



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import contextlib
try:
    from unittest import mock
except ImportError:
    import mock
import os

def unset_environ():
    try:
        del os.environ["TD_API_KEY"]
    except KeyError:
        pass
    try:
        del os.environ["TD_API_SERVER"]
    except KeyError:
        pass
    try:
        del os.environ["HTTP_PROXY"]
    except KeyError:
        pass

def make_raw_response(status, body, headers={}):
    response = mock.MagicMock()
    response.status = status
    response.pos = 0
    response.body = body
    def read(size=None):
        if response.pos < len(response.body):
            if size is None:
                s = body[response.pos:]
                response.pos = len(response.body)
            else:
                s = response.body[response.pos:response.pos+size]
                response.pos += size
            return s
        else:
            return b""
    response.read.side_effect = read
    return response

def make_response(*args, **kwargs):
    response = make_raw_response(*args, **kwargs)
    return contextlib.closing(response)



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_bulk_import():
    client = mock.MagicMock()
    bulk_import = model.BulkImport(client, name="name", database="database", table="table", status="status", upload_frozen="upload_frozen", job_id="job_id", valid_records="valid_records", error_records="error_records", valid_parts="valid_parts", error_parts="error_parts")
    assert bulk_import.name == "name"
    assert bulk_import.database == "database"
    assert bulk_import.table == "table"
    assert bulk_import.status == "status"
    assert bulk_import.job_id == "job_id"
    assert bulk_import.valid_records == "valid_records"
    assert bulk_import.error_records == "error_records"
    assert bulk_import.valid_parts == "valid_parts"
    assert bulk_import.error_parts == "error_parts"
    assert bulk_import.upload_frozen == "upload_frozen"



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

class Model(object):
    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        """
        Returns: a :class:`tdclient.Client` instance
        """
        return self._client



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock

from tdclient import model
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_scheduled_job():
    client = mock.MagicMock()
    schedule = model.Schedule(client, "name", "cron", "query", database="database", result_url="result_url", timezone="timezone", delay="delay", next_time="next_time", priority="priority", retry_limit="retry_limit", org_name="org_name")
    assert schedule.name == "name"
    assert schedule.cron == "cron"
    assert schedule.query == "query"
    assert schedule.database == "database"
    assert schedule.result_url == "result_url"
    assert schedule.timezone == "timezone"
    assert schedule.delay == "delay"
    assert schedule.priority == "priority"
    assert schedule.retry_limit == "retry_limit"
    assert schedule.org_name == "org_name"

def test_schedule():
    pass



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

try:
    from unittest import mock
except ImportError:
    import mock
import pytest

from tdclient import api
from tdclient.test.test_helper import *

def setup_function(function):
    unset_environ()

def test_show_account_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "account":{
                "id":1,
                "plan":0,
                "storage_size":1024,
                "guaranteed_cores":0,
                "maximum_cores":4,
                "created_at":"1970-01-01"
            }
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    access_controls = td.show_account()
    td.get.assert_called_with("/v3/account/show")

def test_show_account_failure():
    td = api.API("APIKEY")
    td.get = mock.MagicMock(return_value=make_response(500, b"error"))
    with pytest.raises(api.APIError) as error:
        td.show_account()
    assert error.value.args == ("500: Show account failed: error",)

def test_core_utilization_success():
    td = api.API("APIKEY")
    # TODO: should be replaced by wire dump
    body = b"""
        {
            "from": "2015-01-03 10:29:34 UTC",
            "to": "2015-01-03 10:30:29 UTC",
            "interval": "1",
            "history": ""
        }
    """
    td.get = mock.MagicMock(return_value=make_response(200, body))
    access_controls = td.account_core_utilization(0, 3)
    td.get.assert_called_with("/v3/account/core_utilization", {"from": "0", "to": "3"})



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from tdclient.access_control_model import AccessControl
from tdclient.account_model import Account
from tdclient.bulk_import_model import BulkImport
from tdclient.database_model import Database
from tdclient.job_model import Schema, Job
from tdclient.result_model import Result
from tdclient.schedule_model import ScheduledJob, Schedule
from tdclient.table_model import Table
from tdclient.user_model import User



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import time

from tdclient._model import Model

class Schema(object):
    """Schema of a database table on Treasure Data Service
    """

    class Field(object):
        def __init__(self, name, type):
            self._name = name
            self._type = type

        @property
        def name(self):
            """
            TODO: add docstring
            """
            return self._name

        @property
        def type(self):
            """
            TODO: add docstring
            """
            return self._type

    def __init__(self, fields=[]):
        self._fields = fields

    @property
    def fields(self):
        """
        TODO: add docstring
        """
        return self._fields

    def add_field(self, name, type):
        """
        TODO: add docstring
        """
        self._fields.append(Field(name, type))

class Job(Model):
    """Job on Treasure Data Service
    """

    STATUS_QUEUED = "queued"
    STATUS_BOOTING = "booting"
    STATUS_RUNNING = "running"
    STATUS_SUCCESS = "success"
    STATUS_ERROR = "error"
    STATUS_KILLED = "killed"
    FINISHED_STATUS = [STATUS_SUCCESS, STATUS_ERROR, STATUS_KILLED]

    JOB_PRIORITY = {
        -2: "VERY LOW",
        -1: "LOW",
        0: "NORMAL",
        1: "HIGH",
        2: "VERY HIGH",
    }

    def __init__(self, client, job_id, type, query, status=None, url=None, debug=None, start_at=None, end_at=None, cpu_time=None,
                 result_size=None, result=None, result_url=None, hive_result_schema=None, priority=None, retry_limit=None,
                 org_name=None, db_name=None):
        super(Job, self).__init__(client)
        self._job_id = job_id
        self._type = type
        self._url = url
        self._query = query
        self._status = status
        self._debug = debug
        self._start_at = start_at
        self._end_at = end_at
        self._cpu_time = cpu_time
        self._result_size = result_size
        self._result = result
        self._result_url = result_url
        self._hive_result_schema = hive_result_schema
        self._priority = priority
        self._retry_limit = retry_limit
        self._org_name = org_name
        self._db_name = db_name

    @property
    def id(self):
        """
        Returns: a string represents the identifier of the job
        """
        return self._job_id

    @property
    def job_id(self):
        """
        Returns: a string represents the identifier of the job
        """
        return self._job_id

    @property
    def type(self):
        """
        Returns: a string represents the engine type of the job (e.g. "hive", "presto", etc.)
        """
        return self._type

    @property
    def result_url(self):
        """
        Returns: a string of URL of the result on Treasure Data Service
        """
        return self._result_url

    @property
    def priority(self):
        """
        Returns: a string represents the priority of the job (e.g. "NORMAL", "HIGH", etc.)
        """
        if self._priority in self.JOB_PRIORITY:
            return self.JOB_PRIORITY[self._priority]
        else:
            # just convert the value in string
            return str(self._priority)

    @property
    def retry_limit(self):
        """
        TODO: add docstring
        """
        return self._retry_limit

    @property
    def org_name(self):
        """
        Returns: organization name
        """
        return self._org_name

    @property
    def db_name(self):
        """
        Returns: a string represents the name of a database that job is running on
        """
        return self._db_name

    def wait(self, timeout=None):
        """Sleep until the job has been finished

        Params:
            timeout (int): Timeout in seconds. No timeout by default.
        """
        started_at = time.time()
        while not self.finished():
            if timeout is None or abs(time.time() - started_at) < timeout:
                time.sleep(1) # TODO: configurable
            else:
                raise RuntimeError("timeout") # TODO: throw proper error

    def kill(self):
        """Kill the job

        Returns: a string represents the status of killed job ("queued", "running")
        """
        return self._client.kill(self.job_id)

    @property
    def query(self):
        """
        Returns: a string represents the query string of the job
        """
        return self._query

    def status(self):
        """
        Returns: a string represents the status of the job ("success", "error", "killed", "queued", "running")
        """
        if self._query is not None and not self.finished():
            self._update_status()
        return self._status

    @property
    def url(self):
        """
        Returns: a string of URL of the job on Treasure Data Service
        """
        return self._url

    def result(self):
        """
        Returns: an iterator of rows in result set
        """
        if not self.finished():
            raise ValueError("result is not ready")
        else:
            if self._result is None:
                for row in self._client.job_result_each(self._job_id):
                    yield row
            else:
                for row in self._result:
                    yield row

    def result_format(self, format):
        """
        Params:
            format (str): output format of result set

        Returns: an iterator of rows in result set
        """
        if not self.finished():
            raise ValueError("result is not ready")
        else:
            if self._result is None:
                for row in self._client.job_result_format_each(self._job_id, format):
                    yield row
            else:
                for row in self._result:
                    yield row

    def finished(self):
        """
        Returns: `True` if the job has been finished in success, error or killed
        """
        self._update_progress()
        return self._status in self.FINISHED_STATUS

    def success(self):
        """
        Returns: `True` if the job has been finished in success
        """
        self._update_progress()
        return self._status == self.STATUS_SUCCESS

    def error(self):
        """
        Returns: `True` if the job has been finished in error
        """
        self._update_progress()
        return self._status == self.STATUS_ERROR

    def killed(self):
        """
        Returns: `True` if the job has been finished in killed
        """
        self._update_progress()
        return self._status == self.STATUS_KILLED

    def queued(self):
        """
        Returns: `True` if the job is queued
        """
        self._update_progress()
        return self._status == self.STATUS_QUEUED

    def running(self):
        """
        Returns: `True` if the job is running
        """
        self._update_progress()
        return self._status == self.STATUS_RUNNING

    def _update_progress(self):
        if self._status not in self.FINISHED_STATUS:
            self.__update_progress()

    def __update_progress(self):
        self._status = self._client.job_status(self._job_id)

    def _update_status(self):
        job = self._client.api.show_job(self._job_id)
        self._query = job.get("query")
        self._status = job.get("status")
        self._url = job.get("url")
        self._debug = job.get("debug")
        self._start_at = job.get("start_at")
        self._end_at = job.get("end_at")
        self._cpu_time = job.get("cpu_time")
        self._result_size = job.get("result_size")
        self._result_url = job.get("result_url")
        self._hive_result_schema = job.get("hive_result_schema")
        self._priority = job.get("priority")
        self._retry_limit = job.get("retry_limit")
        self._db_name = job.get("db_name")



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import os

def where():
    certs = [
        "/etc/ssl/certs/ca-certificates.crt", # debian
        "/etc/ssl/certs/ca-bundle.crt", # redhat
    ]
    for cert in certs:
        if os.path.exists(cert):
            return cert
    return None
